# ReadMe

## USD Dependencies & Changes

### All USD Versions

To ensure FnUsdAbstraction has access to all the methods and classes it requires
to successfully bind the required USD Libraries, the changes required are
listed below.

USD must be built with the following features enabled:
 - usd-imaging
 - python

### USD < 22.08:

A patch was introduced in USD 22.08 to allow for adding to the NdrRegistry at
runtime.  This is a requirement for some of FnUsdAbstraction. We have enabled a
variable `PXR_SDR_DISCOVER_PATCHED` which should be defined to `ON` for CMake if
the patch has been backported. Both Nuke & Katana's USD packages have this patch
backported. https://github.com/PixarAnimationStudios/USD/pull/1810

## Dependencies && CMake Arguments

The same Python version as the application must be used.

We provide a simplified CMakeLists.txt with this source code which should help
starting out a build process.  You may wish to modify this if your third party
dependencies, like OpenSubDiv are not set up as CMake Targets.

- `FdkBase_DIR` : Where the FdkBaseConfig.cmake can be found using
  `find_package`.
- `OpenSubdiv_DIR`: Where OpenSubDivConfig.cmake can be found, or the root to a
standard OpenSubdiv installation. Check the CMakeLists.txt file for how this
tries to search for the required libraries. If using USD's default install, pass
the same as `pxr_DIR`.
- `pxr_DIR` - path to the root of your custom USD installation.
- `CMAKE_INSTALL_PREFIX` - prefix to where FnUsdAbstraction will be installed.

### Katana specifics

- `KATANA_ROOT` must be defined as the root of the Katana install you wish to
build against. It is used by `FdkBaseConfig.cmake` as the root for where the
library and headers are installed.

### Example Katana FnUsdAbstraction build scripts

#### Linux

```
rm -rf build && mkdir build && cd build
cmake .. -G Ninja \
    -DCMAKE_BUILD_TYPE="Release" \
    -Dpxr_DIR=/workspace/USD_2208_py39 \
    -DFdkBase_DIR=/workspace/Katana_6.0/plugin_apis/cmake \
    -DKATANA_ROOT=/workspace/Katana_6.0 \
    -DOpenSubdiv_DIR=/workspace/USD_2208_py39 \
    -DCMAKE_INSTALL_PREFIX=/workspace/FnUsdAbstraction_2208
cmake --build . --config Release
```

#### Windows

```
mkdir build && cd build
cmake .. -G Ninja ^
    -DCMAKE_BUILD_TYPE="Release" ^
    -DHAVE_SNPRINTF=True ^
    -Dpxr_DIR=C:\USD_2208_py39 ^
    -DFdkBase_DIR=C:\Katana_6.0\plugin_apis\cmake ^
    -DKATANA_ROOT=C:\Katana_6.0\ ^
    -DOpenSubdiv_DIR=C:\USD_2208_py39 ^
    -DCMAKE_INSTALL_PREFIX="C:\FnUsdAbstraction_2208"
cmake --build . --config Release
cd ..
```

## Running your custom FnUsdAbstraction in Katana

To run Katana with your own custom FnUsdAbstraction you must use the following environment variables.

The main environment variable to disable Katana's internal FnUsdAbstraction library is `KATANA_USD_ABSTRACTION_DISABLED`. Otherwise, ensure all dependencies are added. Below is an example
after using USD's `build_script.py`.

### Linux

```
export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:/workspace/USD_2208_py39/:/workspace/FnUsdAbstraction_2208/Libs/
export PYTHONPATH=$PYTHONPATH:/workspace/USD_2208_py39/lib/python
export KATANA_USD_ABSTRACTION_DISABLED=1
```

### Windows

Windows requires the additional `usg_LIB_DIRS` because directories on `PATH` are no longer added to
the dll_directories for Python 3.8+. Therefore we manually add any library under `usg_LIB_DIRS` when
the Python bindings are loading its dependencies, which are FnUsdAbstraction and USD by default.

```
set usg_LIB_DIRS=C:\\USD_2208_py39\\lib;C:\\FnUsdAbstraction_2208\\Libs
set PATH=%PATH%;C:\\USD_2208_py39\\lib;
set PYTHONPATH=%PYTHONPATH%;C:\USD_2208_py39\\lib\\python
set KATANA_USD_ABSTRACTION_DISABLED=1
```
