// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#ifndef USG_PRIVATE_MEMORYPOOLWRAPPERIMPL_H
#define USG_PRIVATE_MEMORYPOOLWRAPPERIMPL_H

#include "fdk/base/MemoryPool.h"

namespace usg {
  //! Convenience class to help create MemoryPool-based wrapers for USD classes.
  template<class UsdType, class UsgType, class UsgHandle>
  class MemoryPoolWrapperImpl
  {
  public:
    using MemoryPoolHandle = typename fdk::MemoryPool<UsdType>::Handle;

    MemoryPoolWrapperImpl() = default;

    //! Allocates and constructs an object in the MemoryPool and returns a
    //! handle for it.
    template<typename... Args>
    MemoryPoolHandle allocate(Args&&... args) { return _pool.allocate(std::forward<Args>(args)...); }

    //! Allocates and constructs a copy of the object that handle refers to in
    //! the MemoryPool and returns a handle for it.
    MemoryPoolHandle copyAllocate(const UsgHandle& handle)
    {
      return _pool.allocate(reinterpret_cast<const UsdType&>(handle));
    }

    //! Allocates and constructs a copy of the object that handle refers to in
    //! the MemoryPool and returns a handle for it.
    MemoryPoolHandle copyAllocate(const MemoryPoolHandle& h)
    {
      return _pool.allocate(_pool[h]);
    }

    //! Returns the UsgHandle for the object referred to by MemoryPoolHandle.
    UsgHandle& getWrappedObject(const MemoryPoolHandle& h)
    {
      return reinterpret_cast<UsgHandle&>(_pool[h]);
    }

    //! Creates a Usg object wrapping the given Usd obect.
    UsgType wrap(const UsdType& t)
    {
      return UsgType(reinterpret_cast<const UsgHandle&>(t));
    }

    //! Returns the Usd object in referred to by the MemoryPoolHandle.
    const UsdType& unwrap(const MemoryPoolHandle& h)
    {
      return _pool[h];
    }

    //! Frees the MemoryPool allocation for the object.
    void free(MemoryPoolHandle h) { _pool.free(h); }

  private:
    MemoryPoolWrapperImpl(const MemoryPoolHandle&) = delete;
    MemoryPoolWrapperImpl& operator=(const MemoryPoolWrapperImpl&) = delete;

    fdk::MemoryPool<UsdType> _pool;
  };
}
#endif
