// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#ifndef USG_PRIVATE_USDIMPL_H
#define USG_PRIVATE_USDIMPL_H

#include <cassert>

#include "fdk/base/Hash.h"

#include "usg/base/ArrayTypes.h"

#include "usg/geom/Layer.h"
#include "usg/geom/Stage.h"
#include "usg/geom/Relationship.h"
#include "usg/geom/ShaderDesc.h"
#include "usg/geom/ShaderProperty.h"

#include "usg/private/UsdAttrImpl.h"

#include "pxr/pxr.h"
#include "pxr/usd/sdf/layer.h"
#include "pxr/usd/sdf/listOp.h"
#include "pxr/usd/sdr/shaderNode.h"
#include "pxr/usd/sdr/shaderProperty.h"
#include "pxr/usd/sdf/variantSetSpec.h"
#include "pxr/usd/sdf/variantSpec.h"
#include "pxr/usd/usd/payloads.h"
#include "pxr/usd/usd/prim.h"
#include "pxr/usd/usd/primRange.h"
#include "pxr/usd/usd/stage.h"
#include "pxr/usd/usd/tokens.h"
#include "pxr/usd/usd/notice.h"
#include "pxr/usd/usd/variantSets.h"
#include "pxr/usd/usdGeom/mesh.h"
#include "pxr/usd/usdGeom/bboxCache.h"
#include "pxr/usd/usdGeom/xformCache.h"

#include <atomic>

namespace usg
{

/*
  enum SdfSpecType
  {
      // The unknown type has a value of 0 so that SdfSpecType() is unknown.
      SdfSpecTypeUnknown = 0,

      // Real concrete types
      SdfSpecTypeAttribute,
      SdfSpecTypeConnection,
      SdfSpecTypeExpression,
      SdfSpecTypeMapper,
      SdfSpecTypeMapperArg,
      SdfSpecTypePrim,
      SdfSpecTypePseudoRoot,
      SdfSpecTypeRelationship,
      SdfSpecTypeRelationshipTarget,
      SdfSpecTypeVariant,
      SdfSpecTypeVariantSet,
  };

  enum SdfSpecifier
  {
      SdfSpecifierDef,
      SdfSpecifierOver,
      SdfSpecifierClass,
  };

  enum SdfPermission
  {
      SdfPermissionPublic,         
      SdfPermissionPrivate,        
  };

  SdfVariability:
    An enum that identifies variability types for attributes.
    Variability indicates whether the attribute may vary over time and
    value coordinates, and if its value comes through authoring or
    or from its owner.

    SdfVariabilityVarying: Varying attributes may be directly 
      authored, animated and affected on by Actions.  They are the 
      most flexible.

    SdfVariabilityUniform:  Uniform attributes may be authored 
      only with non-animated values (default values).  They cannot 
      be affected by Actions, but they can be connected to other 
      Uniform attributes.
  
  enum SdfVariability
  {
      SdfVariabilityVarying,
      SdfVariabilityUniform,
  };

*/

extern std::string empty_string;
extern Token       empty_token;
extern Path        empty_geompath;
extern AssetPath   empty_assetpath;


//--------------------------------------------------------------------------------------------------


extern std::string g_preedit_cache_id;
extern std::string g_preedit_cache_hash;


/*! Retrieve the current cache layer metadata values, returning true on success.

    This metdata dictionary is carrying the cache information set by the GeomOp
    management code.
*/
bool getPreEditCacheInfoFromDict(const PXR_NS::VtDictionary& dict,
                                 std::string&                cache_layer_id,
                                 uint64_t&                   cache_layer_hash);


//! A helper function to retrieve a PXR_NS::TfToken from a usg::Token
inline PXR_NS::TfToken toTfToken(const Token& token)
{
  return reinterpret_cast<const PXR_NS::TfToken&>(token.getToken(0/*TODO remove this*/));
}

//! A helper function to retrieve a usg::Token from a PXR_NS::TfToken
inline Token fromTfToken(const PXR_NS::TfToken& token)
{
  return Token(reinterpret_cast<const usg::Token::Handle&>(token));
}

//! A helper function to retrieve a PXR_NS::SdfPath from a usg::Path
inline PXR_NS::SdfPath toSdfPath(const Path& path)
{
  return reinterpret_cast<const PXR_NS::SdfPath&>(path.getPath(0/*TODO remove this*/));
}

//! A helper function to retrieve a usg::Path from a PXR_NS::SdfPath
inline Path fromSdfPath(const PXR_NS::SdfPath& path)
{
  return Path(reinterpret_cast<const usg::Path::Handle&>(path));
}

//! A helper function to retrieve a PXR_NS::SdfAssetPath from a usg::AssetPath
inline PXR_NS::SdfAssetPath toSdfAssetPath(const AssetPath& path)
{
  return reinterpret_cast<const PXR_NS::SdfAssetPath&>(path.getAssetPath(0/*TODO remove this*/));
}

//! A helper function to retrieve a usg::AssetPath from a PXR_NS::SdfAssetPath
inline AssetPath fromSdfAssetPath(const PXR_NS::SdfAssetPath& path)
{
  return AssetPath(reinterpret_cast<const usg::AssetPath::Handle&>(path));
}

//--------------------------------------------------------------------------------------------------


// Once registered with TfNotice this will be called whenever
// there's Sdf edits are performed.
class StageChangeNotice : public PXR_NS::TfWeakBase
{
public:
  StageChangeNotice(Stage::Impl* stageImpl);
  ~StageChangeNotice();

  void _callback(const PXR_NS::UsdNotice::ObjectsChanged& notice,
                 const PXR_NS::UsdStageWeakPtr&           sender);

private:
  Stage::Impl*          _stageImpl;
  PXR_NS::TfNotice::Key _noticeKey;

};

//---------------------------------------

class Stage::Impl
{
public:
  PXR_NS::UsdStageRefPtr  stage;          //!<
  fdk::Hash               compose_state;  //!< Last assigned compose state hash
  fdk::Hash               process_state;  //!< Last assigned process state hash
  fdk::Hash               edit_state;     //!< Last assigned edit state hash
  fdk::Hash               session_state;  //!< Last assigned session state hash

  std::unique_ptr<StageChangeNotice> changeNotifier;

  std::unordered_map<fdk::Hash, PXR_NS::UsdGeomBBoxCache> world_bounds;

public:
  //Impl() : stage_change_tracker(nullptr) {}

  Impl(const Stage::Impl& b) :
    stage(b.stage),
    compose_state(b.compose_state),
    process_state(b.process_state),
    edit_state(b.edit_state),
    session_state(b.session_state),
    changeNotifier(new StageChangeNotice(this))
  {
    //
  }

  Impl(const PXR_NS::UsdStageRefPtr& s) :
    stage(s),
    changeNotifier(new StageChangeNotice(this))
  {
    //
  }
};

//---------------------------------------

inline
StageChangeNotice::StageChangeNotice(Stage::Impl* stageImpl) :
  _stageImpl(stageImpl)
{
  _noticeKey = PXR_NS::TfNotice::Register(PXR_NS::TfCreateWeakPtr(this),
                                          &StageChangeNotice::_callback,
                                          stageImpl->stage);
  //std::cout << "StageChangeNotice::ctor(" << this << ")";
  //std::cout << " stage='" << _stageImpl->stage->GetRootLayer()->GetIdentifier() << "'" << std::endl;
}

inline
StageChangeNotice::~StageChangeNotice()
{
  //std::cout << "StageChangeNotice::dtor(" << this << ")";
  //std::cout << " stage='" << _stageImpl->stage->GetRootLayer()->GetIdentifier() << "'" << std::endl;
  PXR_NS::TfNotice::Revoke(_noticeKey);
}

inline void
StageChangeNotice::_callback(const PXR_NS::UsdNotice::ObjectsChanged& notice,
                             const PXR_NS::UsdStageWeakPtr&           sender)
{
  if (!sender) {
    return;
  }

#if 0
std::cout << "StageChangeNotice::_callback():";
std::cout << " stage='" << sender->GetRootLayer()->GetIdentifier() << "'";
std::cout << ", compose_state=0x" << _stageImpl->compose_state;
std::cout << ", process_state=0x" << _stageImpl->process_state;
std::cout << ", session_state=0x" << _stageImpl->session_state;
std::cout << std::endl;
#endif

}


//--------------------------------------------------------------------------------------------------


class Layer::Impl
{
public:
  PXR_NS::SdfLayerRefPtr  layer;          //!< 
  PXR_NS::SdfChangeBlock* changeBlock = nullptr;
  std::atomic_int         blockCount = 0;


public:
  //Impl() {}

  Impl(const Layer::Impl& b) : layer(b.layer) {}

  Impl(      PXR_NS::SdfLayerRefPtr& l) : layer(l) {}
  Impl(const PXR_NS::SdfLayerRefPtr& l) : layer(l) {}
};


//--------------------------------------------------------------------------------------------------

class Prim::Impl
{
public:
  PXR_NS::SdfPrimSpecHandle  sdfprim;
  PXR_NS::UsdPrim            usdprim;


public:
  Impl() {}
  Impl(Prim::Impl&& b) : sdfprim(std::move(b.sdfprim)), usdprim(std::move(b.usdprim)) {}
  Impl(const Prim::Impl& b) : sdfprim(b.sdfprim), usdprim(b.usdprim) {}

  //! Copy ctor
  Impl(const PXR_NS::SdfPrimSpecHandle& p) : sdfprim(p) {}
  Impl(const PXR_NS::UsdPrim&           p) : usdprim(p) {}
  //! Move ctor
  Impl(PXR_NS::SdfPrimSpecHandle&& p) : sdfprim(std::move(p)) {}
  Impl(PXR_NS::UsdPrim&&           p) : usdprim(std::move(p)) {}

  //--------------------------------------------------------------------

  bool isValid()    const { return (sdfprim  ||  usdprim); }
  bool isSdfPrim()  const { return (sdfprim  && !usdprim); }
  bool isUsdPrim()  const { return (!sdfprim &&  usdprim); }
  bool isEditable() const { return isUsdPrim(); }

  operator bool()   const { return this->isValid(); }

  bool operator == (const Prim::Impl& b) const { return (sdfprim == b.sdfprim && usdprim == b.usdprim); }
  bool operator < (const Prim::Impl& b) const { return (sdfprim ? (sdfprim < b.sdfprim) : (usdprim < b.usdprim)); }

  //--------------------------------------------------------------------

  const std::string& getName() const;
  Path           getPath() const;
  const Token        getTypeName() const;

  void setActive(bool active);
  bool isActive() const;

  bool isDefined() const;

  bool hasPayload() const;
  bool isLoaded() const;

  void load() const;
  void unload() const;

  void setKind(const Token& kind);
  Token getKind() const;

  void setInstanceable(bool instanceable);
  bool getInstanceable() const;

  bool addAppliedSchema(const Token& schemaName);

  void setCustomData(const std::string& key, const Value& value);
  bool getCustomData(const std::string& key, Value& value) const;
  Value getCustomData(const std::string& key) const;
  bool hasCustomData(const std::string& key) const;
  void removeCustomData(const std::string& key);

  //--------------------------------------------------------------------

  //! Create a named Attr with this prim as a parent.
  Attribute createAttr(const PXR_NS::TfToken& name,
                       Value::Type            type,
                       TimeVariability        variability_hint = TimeVariability::Uniform,
                       bool                   is_custom = false);
  Attribute createAttr(const Token&    name,
                       Value::Type     type,
                       TimeVariability variability_hint = TimeVariability::Uniform,
                       bool            is_custom = false);

  //! Retrieve a named Attribute.
  bool getAttr(const PXR_NS::TfToken& name,
               Attribute&              attr);
  bool getAttr(const Token& name,
               Attribute&   attr);


  //! Set an non-array attribute. Returns true on success.
  template <typename usgT, typename usdT>
  Attribute::Impl setAttr(const PXR_NS::TfToken& name,
                          Value::Type       type,
                          TimeVariability        variability_hint,
                          const usgT&            value,
                          const fdk::TimeValue&  time);

  template <typename usgT, typename usdT>
  Attribute::Impl setAttr(const Token&          name,
                          Value::Type           type,
                          TimeVariability       variability_hint,
                          const usgT&           value,
                          const fdk::TimeValue& time);

  //! Set an array attribute. Returns true on success.
  template <typename usgT, typename usdT>
  Attribute::Impl setArrayAttr(const PXR_NS::TfToken&   name,
                               Value::Type              type,
                               TimeVariability          variability_hint,
                               const std::vector<usgT>& value_list,
                               const fdk::TimeValue&    time);


  //--------------------------------------------------------------------

  Relationship createRelationship(const PXR_NS::TfToken& name,
                                  const Path&            targetPath = Path(),
                                  bool                   is_custom = false);

  //! Retrieve a named Relationship.
  bool getRelationship(const PXR_NS::TfToken& name,
                       Relationship&          rel);

  void addReference(const std::string& identifier, const Path& path, ListPosition position);
  void addPayload(const std::string& identifier, const Path& path, ListPosition position);
  void addVariantSet(const std::string& name, ListPosition position);
  void addVariant(const std::string& variantSetName, const std::string& variantName);
  void setVariantSelection(const std::string& variantName, const std::string& selection);
  std::vector<std::string> getVariantSets() const;
  std::vector<std::string> getVariants(const std::string& variantSet) const;

};


//--------------------------------------------------------------------------------------------------

// Implemented in Value.cpp

//!
const usg::ValueTypeInfo&        getValueTypeInfo(Value::Type type);

//! Get the corresponding SdfValueTypeName for a Value::Type.
const PXR_NS::SdfValueTypeName&  getSdfValueTypeName(Value::Type type);
//! Utility to get Value::Type name from enum, as a TfToken.
const PXR_NS::TfToken&           getUsgTypeNameAsToken(Value::Type type);

//! Get the corresponding Value::Type for a SdfValueTypeName, TfType, or ambiguous TfToken.
Value::Type             getUsgValueType(const PXR_NS::SdfValueTypeName& sdf_typename);
Value::Type             getUsgValueType(const PXR_NS::TfType&           tftype);

inline Value::Type      getUsgValueType(const PXR_NS::TfToken& arbitrary_typename)
{
  return Value::getTypeFromName(arbitrary_typename.GetString());
}

//--------------------------------------------------------------------------------------------------


class Value::Impl
{
public:
  PXR_NS::VtValue  vtvalue;
  Value::Type      valuetype;  //!< Local copy for quick Usg-context type testing


public:
  Impl();
  Impl(const Value::Impl&);

  //! Copy ctor
  Impl(const PXR_NS::VtValue&);
  //! Move ctor
  Impl(PXR_NS::VtValue&&);

  //--------------------------------------------------------------------

  bool   isValid() const { return (valuetype != Value::InvalidType); }

  bool   isArray() const { return (vtvalue.IsArrayValued()); }

  //! Returns 0 if not an array type!
  size_t arraySize() const { return vtvalue.GetArraySize(); }

  //! Returns 1 if not an array type, otherwise the array size.
  size_t getSize() const { return (vtvalue.IsArrayValued()) ? vtvalue.GetArraySize() : 1; }

};

inline PXR_NS::VtValue toVtValue(const Value& value)
{
  return value.impl()->vtvalue;
}

inline Value fromVtValue(const PXR_NS::VtValue& value)
{
  return Value(Value::Impl(value));
}


//--------------------------------------------------------------------------------------------------


class Attribute::Impl
{
public:
  PXR_NS::SdfAttributeSpecHandle sdfattr;
  PXR_NS::UsdAttribute           usdattr;
  //
  Value::Type                    valuetype;  //!< Local copy for quick Usg-context type testing
  bool                           is_array;   //!< Local copy for quick testing
  size_t                         size;       //!< Local copy for quick testing


public:
  Impl();
  Impl(const Attribute::Impl&); // copy contents
  Impl(Attribute::Impl&&); // move contents

  //! Copy ctor
  Impl(const PXR_NS::SdfAttributeSpecHandle&);
  Impl(const PXR_NS::UsdAttribute&);
  //! Move ctor
  Impl(PXR_NS::SdfAttributeSpecHandle&&);
  Impl(PXR_NS::UsdAttribute&&);

  //--------------------------------------------------------------------

  bool isValid()    const { return (valuetype != Value::InvalidType); }
  bool isSdf()      const { return (sdfattr  && !usdattr); }
  bool isUsd()      const { return (!sdfattr &&  usdattr); }
  bool isDefined()  const { return (sdfattr || (usdattr && usdattr.IsDefined())); }
  bool isAuthored() const { return (usdattr && usdattr.IsAuthored()); }
  bool isArrayType() const { return is_array; }

  operator bool()   const { return this->isValid(); }

  const std::string& getName() const;
  const std::string& getPath() const;

  //--------------------------------------------------------------------

  //!
  template <typename usgT, typename usdT>
  bool set(const usgT&           value,
           const fdk::TimeValue& time);

  template <typename usgT, typename usdT>
  bool setArray(const std::vector<usgT>& value_list,
                const fdk::TimeValue&    time);

  //!
  bool setVtValue(const PXR_NS::VtValue& vt_value,
                  const fdk::TimeValue&  time) const;

  //!
  template <typename usdT>
  bool setUsd(const usdT&           value,
              const fdk::TimeValue& time);

  template <typename usdT>
  bool setUsdArray(const PXR_NS::VtArray<usdT>& value_array,
                   const fdk::TimeValue&        time);


  //! Remove all authored values, including the default one.
  void clear();
  //! Removes the default value, leaving time samples unaffected.
  void clearDefault();
  //! Remove all time samples from the attribute, leaving the default value unaffected.
  void clearAllTimeSamples();
  //! Remove a specific time sample, ignoring default value.
  void clearAtTime(const fdk::TimeValue& time);

  //! Remove all authored values, leaving the attribute as if unauthored.
  void block();


  //!
  //bool getDefaultValue(PXR_NS::VtValue& v) const
  //{
  //  v = sdfattr->GetField(PXR_NS::SdfFieldKeys->Default);
  //  return (!v.IsEmpty());
  //}

  //! Return 1 if not array type, or the size of the array.
  //size_t size() const
  //{
  //  if (!is_array)
  //    return 1;
  //  const PXR_NS::VtValue v = sdfattr->GetField(PXR_NS::SdfFieldKeys->Default);
  //  if (v.IsEmpty())
  //    return 0;
  //  return v.GetArraySize();
  //}

  //! Return true if number of time samples is greater than 1.
  bool         isAnimating() const;

  //! Return the number of time samples set on the attribute.
  size_t       getNumTimeSamples() const;

  //! Fills the array with all time samples set on the attribute, in ascending order. Returns the size of the array.
  size_t       getTimeSamples(fdk::TimeValueList& times) const;

#if 0
  // TODO: don't think we need this:
  bool         getBracketingTimeSamples(double  desiredTime, 
                                        double* lower, 
                                        double* upper, 
                                        bool*   hasTimeSamples) const;
#endif

  //! Return 1 if not array type, or the size of the array, or 0 if not valid.
  size_t       getSize(const fdk::TimeValue& time) const;

  //!
  template <typename usgT, typename usdT>
  const usgT*  getDefaultData() const;

  //!
  template <typename usgT, typename usdT>
  bool         getDefaultData(const usgT*& data,
                              size_t&      size) const;


  //! Return true if \a vt_value was filled in.
  bool         getVtValue(PXR_NS::VtValue&      vt_value,
                          const fdk::TimeValue& time) const;

  //! Retrieve value from attribute, returning 0 on fail or 1 on success. 
  template <typename usgT, typename usdT>
  size_t       get(usgT&                 value,
                   const fdk::TimeValue& time) const;
  //! Retrieve value from attribute, returning 0 on fail or size of array on success. 
  template <typename usgT, typename usdT>
  size_t       getArray(std::vector<usgT>&    value_list,
                        const fdk::TimeValue& time) const;

  template <typename usdT>
  bool         getUsd(usdT&                 value,
                      const fdk::TimeValue& time) const;
  template <typename usdT>
  bool         getUsdArray(PXR_NS::VtArray<usdT>& value_array,
                           const fdk::TimeValue&  time) const;

  //--------------------------------------------------------------------

  //!
  void          getFieldNames(fdk::StringList& field_names) const;
  //! Return true if Attribute has the *non-empty* named field.
  bool          hasField(const PXR_NS::TfToken& field_name) const;
  //! Returns the content of the non-empty named field, otherwise an invalid Value.
  PXR_NS::VtValue  getField(const PXR_NS::TfToken& field_name) const;
  //! Set the contents of a field, creating it if required.
  bool          setField(const PXR_NS::TfToken& field_name,
                         const PXR_NS::VtValue& vtvalue);
  //! Remove the named field from the Attribute.
  bool          deleteField(const PXR_NS::TfToken& field_name);

  //--------------------------------------------------------------------

  bool        addConnectionTo(const Path& source);

  bool        removeConnectionTo(const Path& source);

  void        clearConnections();

  Path getConnection() const;


  //--------------------------------------------------------------------


};


//--------------------------------------------------------------------------------------------------


class Relationship::Impl
{
public:
  PXR_NS::SdfRelationshipSpecHandle sdfrel;
  PXR_NS::UsdRelationship           usdrel;

public:
  Impl() {}
  Impl(const Relationship::Impl& b) : sdfrel(b.sdfrel), usdrel(b.usdrel) {}

  //! Copy ctor
  Impl(const PXR_NS::SdfRelationshipSpecHandle& rel) : sdfrel(rel) {}
  Impl(const PXR_NS::UsdRelationship&           rel) : usdrel(rel) {}
  //! Move ctor
  Impl(PXR_NS::SdfRelationshipSpecHandle&& rel) : sdfrel(std::move(rel)) {}
  Impl(PXR_NS::UsdRelationship&&           rel) : usdrel(std::move(rel)) {}

  //--------------------------------------------------------------------

  bool isSdf()  const { return ( sdfrel && !usdrel); }
  bool isUsd()  const { return (!sdfrel &&  usdrel); }

  bool isValid() const { return (sdfrel || usdrel); }

  const std::string& getName() const;
  const std::string& getPath() const;

  //--------------------------------------------------------------------

#if 0
  //! USD - Make the authoring layer's opinion of the targets list explicit, and set exactly to \p targets.
  bool             setTargets(const SdfPathVector& targets) const;


  //! SDF - Returns the relationship's target path list editor.
  SdfTargetsProxy  getTargetList() const;
  //! USD - Compose this relationship's targets and fill \p targets with the result.
  bool             getTargets(SdfPathVector* targets) const;



  //! SDF - Returns true if the relationship has any target paths.
  bool             hasTargetList() const;
  //! SDF - Clears the list of target paths on this relationship.
  void             clearTargetList() const;
  //! USD - Remove all opinions about the target list from the current edit target.
  bool             clearTargets(bool removeSpec) const;


  //! SDF - Updates the specified target path.
  void             replaceTarget(const SdfPath& oldPath,
                                 const SdfPath& newPath);
  //! USD - 
  bool             addTarget(const SdfPath&  target,
                             UsdListPosition position = UsdListPositionBackOfPrependList) const;


  //! SDF - Removes the specified target path.
  void             removeTarget(const SdfPath& path,
                                bool           preserveTargetOrder = false);
  //! USD - Removes \p target from the list of targets.
  bool             removeTarget(const SdfPath& target) const;

  // Misc:

  //! SDF - Get whether loading the target of this relationship is necessary to load the prim we're attached to
  bool  getNoLoadHint(void) const;
  //! SDF - Set whether loading the target of this relationship is necessary to load the prim we're attached to
  void  setNoLoadHint(bool noload);

  //! USD - Returns true if any target path opinions have been authored. 
  bool  hasAuthoredTargets() const;
  //! USD - Compose this relationship's \em ultimate targets, taking into account "relationship forwarding", and fill \p targets with the result.
  bool  getForwardedTargets(SdfPathVector* targets) const;


#endif

  // Insert the target path.
  bool setTarget(const Path& targetPath)
  {
    if (sdfrel) {
      sdfrel->GetTargetPathList().GetExplicitItems().Insert(0, toSdfPath(targetPath));
      return true;
    }
    else if (usdrel && usdrel.IsValid()) {
      usdrel.AddTarget(toSdfPath(targetPath));
      return true;
    }

    return false;
  }

  //!
  Path getTarget() const
  {
    if (sdfrel) {
      if (sdfrel->GetTargetPathList().GetExplicitItems().size() > 0) {
        const PXR_NS::SdfPath& path = sdfrel->GetTargetPathList().GetExplicitItems()[0];
        return fromSdfPath(path);
      }
    }
    else if (usdrel && usdrel.IsValid()) {
      PXR_NS::SdfPathVector targets;
      usdrel.GetTargets(&targets);
      if (targets.size() > 0) {
        return fromSdfPath(targets[0]);
      }
    }

    return empty_geompath;
  }

  // Insert the target paths.
  bool setTargets(const PathArray& targetPaths)
  {
    if (sdfrel) {
      sdfrel->GetTargetPathList().GetExplicitItems().clear();
      for (auto&& path : targetPaths) {
        sdfrel->GetTargetPathList().GetExplicitItems().push_back(toSdfPath(path));
      }
      return true;
    }
    else if (usdrel && usdrel.IsValid()) {
      usdrel.ClearTargets(true);
      for (auto&& path : targetPaths) {
        usdrel.AddTarget(toSdfPath(path));
      }
      return true;
    }

    return false;
  }

  //!
  PathArray getTargets() const
  {
    PathArray result;
    if (sdfrel) {
      for (auto&& path : sdfrel->GetTargetPathList().GetExplicitItems()) {
        result.push_back(fromSdfPath(path));
      }
    }
    else if (usdrel && usdrel.IsValid()) {
      PXR_NS::SdfPathVector targets;
      usdrel.GetTargets(&targets);
      for (auto&& path : targets) {
        result.push_back(fromSdfPath(path));
      }
    }

    return result;
  }

};


//--------------------------------------------------------------------------------------------------


/*!
*/
class ShaderPropertyHandle
{
public:
  struct SdrTypeParams
  {
    Value::Type sdr_type;     //!< Converted to base Sdr type (Float if original type was Float3, Int if original type was Int[])
    int32_t     elements;     //!<  > 1 if type was converted from a compound type to a array type
    bool        is_array;     //!< If Sdf type is array variant of the sdr_type base (Int[] vs Int)

    SdrTypeParams(Value::Type _sdr_type, int32_t _elements=0, bool _is_array=false) :
      sdr_type(_sdr_type), elements(_elements), is_array(_is_array) {}
  };

  /*! Contains all the bits required to construct a SdrShaderProperty
      in a single blob so it can be filled in and passed by
      mapping methods.
  */
  struct SdrTypeDescription
  {
    Value::Type          usg_type;        //!< Original usg type
    PXR_NS::TfToken      sdr_typename;    //!< May be remapped from Value::Type passed in from ShaderSchema ctor
    size_t               array_size;      //!< May be used by fixed-sized types like float2, float4, etc.
    PXR_NS::VtValue      default_vtvalue; //!< May be re-defined from value passed in from ShaderSchema ctor
    //
    PXR_NS::NdrTokenMap  metadata;         //!< Converted from KeyValueMap
    PXR_NS::NdrTokenMap  hints;            //!< Converted from KeyValueMap
    PXR_NS::NdrOptionVec options;          //!< Converted from StringList
  };


public:
  const PXR_NS::SdrShaderProperty*  sdrrefproperty;     //!< Reference to const property
  PXR_NS::SdrShaderProperty*        sdrovrproperty;     //!< Localized (overriden) property instance
  fdk::TimeValue                    override_time;      //!< Time the override is sampled at
  //
  Value::Type                       valuetype;          //!< Local copy for quick Usg-context type testing


public:
  //! Default ctor leaves both property ptrs null.
  ShaderPropertyHandle() :
    sdrrefproperty(nullptr),
    sdrovrproperty(nullptr),
    override_time(fdk::defaultTimeValue()),
    valuetype(Value::InvalidType)
  {
    //
  }

  /*! Make a reference to a const property.
  */
  ShaderPropertyHandle(const PXR_NS::SdrShaderProperty* constproperty) :
    sdrrefproperty(constproperty),
    sdrovrproperty(nullptr),
    override_time(fdk::defaultTimeValue()),
    valuetype((constproperty) ? getUsgValueType(constproperty->GetType()) : Value::InvalidType)
  {
    //
  }

  /*! Copy ctor copies the property reference and if valid does NOT copy
      the local property data.
  */
  ShaderPropertyHandle(const ShaderPropertyHandle& b) :
    sdrrefproperty(b.sdrrefproperty),
    sdrovrproperty((!b.sdrrefproperty && b.sdrovrproperty) ?
                      new PXR_NS::SdrShaderProperty(b.sdrovrproperty->GetName(),
                                                    b.sdrovrproperty->GetType(),
                                                    b.sdrovrproperty->GetDefaultValue(),
                                                    b.sdrovrproperty->IsOutput(),
                                                    b.sdrovrproperty->GetArraySize(),
                                                    b.sdrovrproperty->GetMetadata(),
                                                    b.sdrovrproperty->GetHints(),
                                                    b.sdrovrproperty->GetOptions()) :
                        nullptr),
    override_time(b.override_time),
    valuetype(b.valuetype)
  {
    //
  }


  /*! Local property ctor, typically for default property ctor from ShaderSchema
      property list.
      The SdrTypeDescription is used as type converter info in a single blob.
  */
  ShaderPropertyHandle(const PXR_NS::TfToken&    name,
                       int                       usg_category,
                       const SdrTypeDescription& sdr_description) :
    sdrrefproperty(nullptr),
    sdrovrproperty(new PXR_NS::SdrShaderProperty(name,
                                                 sdr_description.sdr_typename,
                                                 sdr_description.default_vtvalue,
                                                 (usg_category == usg::OutputProp)/*isOutput*/,
                                                 sdr_description.array_size,
                                                 sdr_description.metadata,
                                                 sdr_description.hints,
                                                 sdr_description.options)),
    override_time(fdk::defaultTimeValue()),
    valuetype(sdr_description.usg_type)
  {
    //
  }


  /*! Local property ctor, typically for an override definition.
      Note- this ctor assumes the calling code has pre-configured the usg local
      category metadata tag in the passed-in metadata!
  */
  ShaderPropertyHandle(const PXR_NS::TfToken&      name,
                       const PXR_NS::TfToken&      usd_typename,
                       const PXR_NS::VtValue&      value,
                       int                         usg_category,
                       size_t                      arraySize,
                       const PXR_NS::NdrTokenMap&  metadata,
                       const PXR_NS::NdrTokenMap&  hints,
                       const PXR_NS::NdrOptionVec& options) :
    sdrrefproperty(nullptr),
    sdrovrproperty(new PXR_NS::SdrShaderProperty(name,
                                                 usd_typename,
                                                 value,
                                                 (usg_category == usg::OutputProp)/*isOutput*/,
                                                 arraySize,
                                                 metadata,
                                                 hints,
                                                 options)),
    override_time(fdk::defaultTimeValue())
  {
    if (sdrovrproperty) {
      valuetype = getUsgValueType(sdrovrproperty->GetType());
    }
  }


  //!
  ~ShaderPropertyHandle()
  {
    if (sdrovrproperty) {
      delete sdrovrproperty;
    }
  }


  //! Returns the const property if assigned otherwise the ptr of the local property.
  const PXR_NS::SdrShaderProperty* getProperty() const
  {
    // TODO: maybe the local should take precedence...?
    return (sdrrefproperty) ? sdrrefproperty : sdrovrproperty;
  }


  bool  isInput()  const { return  isPropertyInput(getProperty()); }
  bool  isLocal()  const { return  isPropertyLocal(getProperty()); }
  bool  isOutput() const { return !isPropertyInput(getProperty()); }


  // Implemented in SchemaProperty.cpp
  static void         usgToSdrMetadata(const fdk::KeyValueMap& usgmetadata,
                                       PXR_NS::NdrTokenMap&    sdrmetadata);
  static void         usgToSdrPropertyArray(const ShaderPropertyArray&       usgproperties,
                                            PXR_NS::NdrPropertyUniquePtrVec& sdrproperties);

  //! Convert commonly-used Value::Types to the correct Sdr type to avoid common propery declaration mistakes.
  static Value::Type   conformUsgValueTypeToSdrValueType(Value::Type usgtype_hint);

  //! Convert commonly-used Value::Types to the correct VtValue TfType based on the type_hint.
  static PXR_NS::VtValue  conformUsgValueToVtValue(const Value& usg_val_in,
                                                   Value::Type  usgtype_hint);

  static PXR_NS::VtValue  conformUsgValueToVtValue(const Value&         usg_val_in,
                                                   const SdrTypeParams& out_type_params);

  //! Add usg local property category tag to metadata.
  static void insertUsgLocalTaginMetadata(PXR_NS::NdrTokenMap& sdrmetadata)
  {
    sdrmetadata.insert({PXR_NS::TfToken("usg_local"), "1"});
  }

  //! Removes any leading 'inputs:' or 'outputs:' prefix from the name. Category can be provided to reinforce this.
  static PXR_NS::TfToken    conformPropNameToCategory(const char* prop_name,
                                                      int         usg_prop_category);

  static SdrTypeDescription convertToNdrValueAndType(const char*             debug_name,
                                                     int                     usg_prop_category,
                                                     Value::Type             usg_type,
                                                     const Value&            usg_value,
                                                     size_t                  array_size,
                                                     const fdk::KeyValueMap& usg_metadata,
                                                     const fdk::KeyValueMap& usg_hints,
                                                     const fdk::StringList&  usg_options);


  //! Returns true if the property is an input and does NOT have a 'usg_local' metadata tag.
  static bool isPropertyInput(const PXR_NS::NdrProperty* prop)
  {
    return (prop && !prop->IsOutput() && !isPropertyLocal(prop->GetMetadata()));
  }

  //! Returns true if the property is an input and DOES have a 'usg_local' metadata tag.
  static bool isPropertyLocal(const PXR_NS::NdrProperty* prop)
  {
    return (prop && !prop->IsOutput() && isPropertyLocal(prop->GetMetadata()));
  }

  //! Returns true if the metadata contains a 'usg_local' metadata tag.
  static bool isPropertyLocal(const PXR_NS::NdrTokenMap& metadata)
  {
    return (metadata.find(PXR_NS::TfToken("usg_local")) != metadata.end());
  }


  //! Debugging
  static void printNdrProperty(const char* prefix, std::ostream&, const PXR_NS::NdrProperty&);
  static void printSdrProperty(const char* prefix, std::ostream& o, const PXR_NS::SdrShaderProperty& p) { printNdrProperty(prefix, o, p); }
  static void printPropertyHandle(const char* prefix, std::ostream& o, const ShaderPropertyHandle& handle) { printSdrProperty(prefix, o, *handle.getProperty()); }

};


//-----------------------------------------------


class ShaderSchemaHandle
{
public:
  PXR_NS::NdrPropertyUniquePtrVec properties;      //!< 
  PXR_NS::NdrTokenMap             metadata;        //!<

public:
  ShaderSchemaHandle(PXR_NS::NdrPropertyUniquePtrVec&& _properties,
                     PXR_NS::NdrTokenMap&&             _metadata) :
    properties(std::move(_properties)),
    metadata(std::move(_metadata))
  {
    //
  }


  //! Debugging
  static void printNdrNode(const char* prefix, std::ostream&, const PXR_NS::NdrNode&);
  static void printSdrShaderNode(const char* prefix, std::ostream&, const PXR_NS::SdrShaderNode&);


};


//-----------------------------------------------


class ShaderDescHandle
{
public:
  const PXR_NS::NdrNode*  ndrnode;          //!< Reference to the const ShaderSchema registry node
  //
  std::string         class_name;           //!<
  std::string         rel_prim_path;        //!< Shader prim path relative to Material prim parent
  ShaderPropertyMap   input_overrides;      //!<
  ShaderConnectionMap input_connections;    //!<

public:
  //!
  ShaderDescHandle(const PXR_NS::NdrNode* node,
                   const char*            prim_path = "") :
    ndrnode(node),
    rel_prim_path(prim_path)
  {
    if (ndrnode) {
      //if (!ndrnode->GetImplementationName().empty())
     //   class_name = ndrnode->GetImplementationName();
      //else
        class_name = ndrnode->GetIdentifier().GetString();
    }
  }


  //!
  bool isValid() const { return (ndrnode != nullptr); }


  //! Delete overrides and connections
  ~ShaderDescHandle()
  {
    for (ShaderPropertyMap::const_iterator it=input_overrides.begin(); it != input_overrides.end(); ++it) {
      delete it->second;
    }
    for (ShaderConnectionMap::const_iterator it=input_connections.begin(); it != input_connections.end(); ++it) {
      delete it->second;
    }
  }


private:
  /*! Disable the copy ctor in favor of an explicit ShaderDesc version
      so we always construct from a SdrShaderNode reference.
  */
  ShaderDescHandle(const ShaderDescHandle& b) = delete;
  ShaderDesc& operator = (const ShaderDesc&) = delete;

};


//--------------------------------------------------------------------------------------------------



/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/
/*                   Inline Function Implementations                   */
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/


inline Value::Impl::Impl()
  : vtvalue()
  , valuetype(Value::InvalidType)
{
  //
}

inline Value::Impl::Impl(const Value::Impl& b)
  : vtvalue(b.vtvalue)
  , valuetype(b.valuetype)
{
  //
}

inline Value::Impl::Impl(const PXR_NS::VtValue& b)
  : vtvalue(b) // copy
  , valuetype(getUsgValueType(b.GetType()/*TfType*/))
{
  //
}

inline Value::Impl::Impl(PXR_NS::VtValue&& b)
  : vtvalue(std::move(b)) // move
{
  // Must get the TfType from the constructed VtValue not the source one (not sure why!!!!!)
  valuetype = getUsgValueType(vtvalue.GetType()/*TfType*/);

  //std::cout << "Value::Impl::ctor(): vtvalue='" << vtvalue.GetTypeName() << "'";
  //std::cout << ", tftype=0x" << std::hex <<  vtvalue.GetType().GetTypeid().hash_code() << std::dec;
  //std::cout << ", valuetype='" << Value::getTypeName(valuetype) << "'";
  //std::cout << std::endl;
}


//----------------------------------------------------------------------

inline Attribute::Impl::Impl()
  : valuetype(Value::InvalidType)
  , is_array(false)
{
  //
}

inline Attribute::Impl::Impl(const Attribute::Impl& b)
  : sdfattr(b.sdfattr)
  , usdattr(b.usdattr)
  , valuetype(b.valuetype)
  , is_array(b.is_array)
{
  //
}

inline Attribute::Impl::Impl(Attribute::Impl&& b)
  : sdfattr(std::move(b.sdfattr))
  , usdattr(std::move(b.usdattr))
  , valuetype(std::move(b.valuetype))
  , is_array(std::move(b.is_array))
{
  //
}

inline Attribute::Impl::Impl(PXR_NS::SdfAttributeSpecHandle&& a)
  : sdfattr(std::move(a))
{
  if (sdfattr && !sdfattr->GetPath().IsEmpty()) {
    valuetype = getUsgValueType(sdfattr->GetTypeName());
    is_array  = sdfattr->GetTypeName().IsArray();
  }
  else {
    valuetype = Value::InvalidType;
    is_array  = false;
  }
}

inline Attribute::Impl::Impl(const PXR_NS::SdfAttributeSpecHandle& a)
  : sdfattr(a)
{
  if (sdfattr && !sdfattr->GetPath().IsEmpty()) {
    valuetype = getUsgValueType(sdfattr->GetTypeName());
    is_array  = sdfattr->GetTypeName().IsArray();
  }
  else {
    valuetype = Value::InvalidType;
    is_array  = false;
  }
}

inline Attribute::Impl::Impl(PXR_NS::UsdAttribute&& a)
  : usdattr(std::move(a))
{
  if (usdattr.IsValid()) {
    valuetype = getUsgValueType(usdattr.GetTypeName());
    is_array  = usdattr.GetTypeName().IsArray();
  }
  else {
    valuetype = Value::InvalidType;
    is_array  = false;
  }
}

inline Attribute::Impl::Impl(const PXR_NS::UsdAttribute& a)
  : usdattr(a)
{
  if (usdattr.IsValid()) {
    valuetype = getUsgValueType(usdattr.GetTypeName());
    is_array  = usdattr.GetTypeName().IsArray();
  }
  else {
    valuetype = Value::InvalidType;
    is_array  = false;
  }
}


//!
inline const std::string& Attribute::Impl::getName() const
{
  if (sdfattr) {
    return sdfattr->GetPath().GetNameToken().GetString();
  }
  else if (usdattr) {
    return usdattr.GetName().GetString();
  }
  return empty_string;
}

//!
inline const std::string& Attribute::Impl::getPath() const
{
  if (sdfattr) {
    return sdfattr->GetPath().GetString();
  }
  else if (usdattr) {
    return usdattr.GetPath().GetString();
  }
  return empty_string;
}


//---------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------


inline bool getPreEditCacheInfoFromDict(const PXR_NS::VtDictionary& dict,
                                        std::string&                cache_layer_id,
                                        uint64_t&                   cache_layer_hash)
{
  cache_layer_id.clear();
  cache_layer_hash = 0;

  PXR_NS::VtDictionary::const_iterator it;

  // Get cache layer id:
  it = dict.find(g_preedit_cache_id);
  if (it == dict.end() || !it->second.IsHolding<std::string>()) {
    return false;
  }
  cache_layer_id = it->second.UncheckedGet<std::string>();

  // Get cache hash:
  it = dict.find(g_preedit_cache_hash);
  if (it == dict.end() || !it->second.IsHolding<uint64_t>()) {
    return false;
  }
  cache_layer_hash = it->second.UncheckedGet<uint64_t>();

  return true;
}



/*  Before applying the edit save the current attribute's value to the pre-edit cache
    if one exists and assigned to the UsdStage.

    We find this cache Layer via the current pre-edit cache metadata stored on the
    UsdStage retrieved via the UsdAttribute.
*/
inline bool setPreEditCacheLayerAttribute(const PXR_NS::UsdAttribute& usdattr,
                                          const fdk::TimeValue&       time)
{
  if (!usdattr || !usdattr.IsValid()) {
    return false; // don't crash..
  }

  // Get the metdata dictionary carrying the cache information set by the GeomOp
  // management code:
  PXR_NS::UsdStageWeakPtr stage = usdattr.GetStage();
  assert(stage);

  std::string stage_preedit_cache_id;
  uint64_t    stage_preedit_cache_hash;
  {
    PXR_NS::VtDictionary dict;
    if (!stage->GetMetadata(PXR_NS::SdfFieldKeys->CustomLayerData, &dict)) {
      std::cerr << "setPreEditCacheLayerAttribute: ********* FAILED TO GET STAGE METADATA DICT *********" << std::endl;
      return false; // don't crash...
    }
    if (!getPreEditCacheInfoFromDict(dict, stage_preedit_cache_id, stage_preedit_cache_hash)) {
      std::cerr << "setPreEditCacheLayerAttribute: ********* FAILED TO RETRIEVE STAGE CACHE METADATA *********" << std::endl;
      return false; // can't find stage metadata!
    }
    //std::cout << "setPreEditCacheLayerAttribute('') stage id='" << stage_preedit_cache_id << "', hash=0x" << std::hex << stage_preedit_cache_hash << std::dec << std::endl;
  }

  if (stage_preedit_cache_id.empty()) {
    //std::cout << "setPreEditCacheLayerAttribute('') NO PREEDIT CACHE time=" << time << ", attr='" << usdattr.GetPath() << "'" << std::endl;
    return true; // no filled preedit cache assigned to Stage yet, skip this cache
  }

  // Find the cache SdfLayer:
  //std::cout << "setPreEditCacheLayerAttribute('" << stage_preedit_cache_id << "')";
  //std::cout << " time=" << time << ", attr='" << usdattr.GetPath() << "'";
  PXR_NS::SdfLayerHandle cache_layer = PXR_NS::SdfLayer::Find(stage_preedit_cache_id);
  if (!cache_layer) {
    std::cerr << " setPreEditCacheLayerAttribute: ********* FAILED TO RETRIEVE PRE-EDIT CACHE LAYER *********" << std::endl;
    return false;
  }

  // Got the layer, compare cache state metadata, which may not be assigned yet:
  std::string preedit_layer_cache_id;
  uint64_t    preedit_layer_cache_hash;
  {
    const PXR_NS::VtDictionary dict = cache_layer->GetCustomLayerData();
    getPreEditCacheInfoFromDict(dict, preedit_layer_cache_id, preedit_layer_cache_hash);
  }
  //std::cout << ", stage_edit_hash=0x" << std::hex << stage_preedit_cache_hash << std::dec;
  //std::cout << ", cache_edit_hash=0x" << std::hex << preedit_layer_cache_hash << std::dec;

  // Compare the values:
  if (stage_preedit_cache_id   == preedit_layer_cache_id &&
      stage_preedit_cache_hash == preedit_layer_cache_hash) {
    //std::cout << "setPreEditCacheLayerAttribute('" << stage_preedit_cache_id << "')";
    //std::cout << " time=" << time << ", attr='" << usdattr.GetPath() << "'";
    //std::cout << " - CLEAN - don't update" << std::endl;
  }
  else {
    //---------------------------------------------------------------------------------
    // Cache state is dirty, save the attribute value.
    //
    // Note that we can't update the cache layer's dictionary now as that will stop other
    // attributes from caching. The cache layer's dictionary is updated by the GeomOp
    // state management code.
    //---------------------------------------------------------------------------------

    // Create/get the prim and attr specs for this attribute in the cache layer:
    PXR_NS::SdfAttributeSpecHandle sdfattr = cache_layer->GetAttributeAtPath(usdattr.GetPath());
    if (!sdfattr) {
      PXR_NS::SdfJustCreatePrimAttributeInLayer(cache_layer,
                                                usdattr.GetPath(),
                                                usdattr.GetTypeName(),
                                                usdattr.GetVariability(),
                                                usdattr.IsCustom());
      sdfattr = cache_layer->GetAttributeAtPath(usdattr.GetPath());
    }
    assert(sdfattr);

    // Save the current attribute value to the cache SdfLayer:
    PXR_NS::VtValue current_val;
    if (!usdattr.Get(&current_val, PXR_NS::UsdTimeCode(time))) {
      std::cerr << "setPreEditCacheLayerAttribute: ********* FAILED TO GET USDATTR VALUE *********" << std::endl;
      return false;
    }

    if (fdk::isDefaultTimeValue(time)) {
      sdfattr->SetField(PXR_NS::SdfFieldKeys->Default, current_val);
    }
    else {
      // Access time samples through SdfLayer:
      assert(sdfattr->GetLayer());
      sdfattr->GetLayer()->SetTimeSample(sdfattr->GetPath(), time, current_val);
    }
    //std::cout << "setPreEditCacheLayerAttribute('" << stage_preedit_cache_id << "')";
    //std::cout << " time=" << time << ", attr='" << usdattr.GetPath() << "'";
    //std::cout << " - DIRTY - save value to cache" << std::endl;
  }

  return true;
}


//---------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------


inline bool Attribute::Impl::setVtValue(const PXR_NS::VtValue& vt_value,
                                       const fdk::TimeValue&  time) const
{
  if (sdfattr) {
    if (fdk::isDefaultTimeValue(time)) {
      sdfattr->SetField(PXR_NS::SdfFieldKeys->Default, vt_value);
      //sdfattr->SetDefaultValue(vt_value);
    }
    else {
      // Access time samples through SdfLayer:
      assert(sdfattr->GetLayer());
      sdfattr->GetLayer()->SetTimeSample(sdfattr->GetPath(), time, vt_value);
    }
    return true;
  }
  else if (usdattr && usdattr.IsValid()) {
    // Before applying the edit save the current attribute's value to the pre-edit cache:
    //setPreEditCacheLayerAttribute(usdattr, time);

    usdattr.Set(vt_value, PXR_NS::UsdTimeCode(time));
  }

  return false;
}

//!
template <typename usdT>
inline bool Attribute::Impl::setUsd(const usdT&           value,
                                   const fdk::TimeValue& time)
{
  // Defining or editing?
  if (sdfattr) {
    // Work around SetDefaultValue not working:
    if (fdk::isDefaultTimeValue(time)) {
      sdfattr->SetField(PXR_NS::SdfFieldKeys->Default, PXR_NS::VtValue(value));//sdfattr->SetDefaultValue(PXR_NS::VtValue(value));
    }
    else {
      // Access time samples through SdfLayer:
      assert(sdfattr->GetLayer());
      sdfattr->GetLayer()->SetTimeSample(sdfattr->GetPath(), time, PXR_NS::VtValue(value));
    }
    return true;
  }
  else if (usdattr && usdattr.IsValid()) {
    // Before applying the edit save the current attribute's value to the pre-edit cache:
    //setPreEditCacheLayerAttribute(usdattr, time);

    usdattr.Set(PXR_NS::VtValue(value), PXR_NS::UsdTimeCode(time));

    return true;
  }

  return false;
}

//!
template <typename usdT>
inline bool Attribute::Impl::setUsdArray(const PXR_NS::VtArray<usdT>& value_array,
                                        const fdk::TimeValue&        time)
{
  // Defining or editing?
  if (sdfattr) {
    // Work around SetDefaultValue not working:
    if (fdk::isDefaultTimeValue(time)) {
      sdfattr->SetField(PXR_NS::SdfFieldKeys->Default, PXR_NS::VtValue(value_array));//sdfattr->SetDefaultValue(vtarray);
    }
    else {
      // Access time samples through SdfLayer:
      assert(sdfattr->GetLayer());
      sdfattr->GetLayer()->SetTimeSample(sdfattr->GetPath(), time, PXR_NS::VtValue(value_array));
    }
    return true;
  }
  else if (usdattr && usdattr.IsValid()) {
    // Before applying the edit save the current attribute's value to the pre-edit cache:
    //setPreEditCacheLayerAttribute(usdattr, time);

    usdattr.Set(PXR_NS::VtValue(value_array), PXR_NS::UsdTimeCode(time));

    return true;
  }

  return false;
}

//---------------------------------------------------------------------------

//!
template <typename usgT, typename usdT>
inline bool Attribute::Impl::set(const usgT&           value,
                                const fdk::TimeValue& time)
{
  //if (sizeof(usgT) != sizeof(usdT))
  //  return false; // don't crash...
  return setUsd(reinterpret_cast<const usdT&>(value), time);
}

// Specialization for string handling
template <>
inline bool Attribute::Impl::set<std::string, PXR_NS::TfToken>(const std::string&    value,
                                                              const fdk::TimeValue& time)
{
  const PXR_NS::TfToken token(value);
  return setUsd(token, time);
}

//-------------------------------------------------

//!
template <typename usgT, typename usdT>
inline bool Attribute::Impl::setArray(const std::vector<usgT>& value_list,
                                     const fdk::TimeValue&    time)
{
  //if (sizeof(usgT) != sizeof(usdT))
  //  return false; // don't crash...

  // This sucks because Usd forces us to copy the VtArray rather than allowing
  // an allocation of memory and then write to that...
  // However I may misunderstand whether or not the pointer in VtArray is moved
  // into the Usd Stage storage rather than copied.
  const std::vector<usdT>& src_list = reinterpret_cast<const std::vector<usdT>&>(value_list);
  const PXR_NS::VtArray<usdT> vtarray(src_list.data(), src_list.data() + src_list.size());

  //std::cout << "        array=[";
  //for (int i=0; i < vtarray.size(); ++i)
  //  std::cout << " " << vtarray[i];
  //std::cout << " ]" << std::endl;

  return setUsdArray(vtarray, time);
}

//!
template <>
inline bool Attribute::Impl::setArray<std::string, PXR_NS::TfToken>(const std::vector<std::string>& value_list,
                                                                   const fdk::TimeValue&           time)
{
  // This sucks because Usd forces us to copy the VtArray rather than allowing
  // an allocation of memory and then write to that...
  // However I may misunderstand whether or not the pointer in VtArray is moved
  // into the Usd Stage storage rather than copied.
  PXR_NS::VtArray<PXR_NS::TfToken> vtarray;
  vtarray.reserve(value_list.size());
  for (size_t i=0; i < value_list.size(); ++i) {
    vtarray.push_back(PXR_NS::TfToken(value_list[i]));
  }
  return setUsdArray(vtarray, time);
}

template <>
inline bool Attribute::Impl::setArray<Token, PXR_NS::TfToken>(const std::vector<Token>& value_list,
                                                              const fdk::TimeValue&     time)
{
  PXR_NS::VtArray<PXR_NS::TfToken> vtarray;
  vtarray.reserve(value_list.size());
  for (size_t i=0; i < value_list.size(); ++i) {
    vtarray.push_back(toTfToken(value_list[i]));
  }
  return setUsdArray(vtarray, time);
}

template <>
inline bool Attribute::Impl::setArray<bool, bool>(const std::vector<bool>& value_list,
                                                  const fdk::TimeValue& time)
{
  const PXR_NS::VtArray<bool> vtarray(value_list.cbegin(), value_list.cend());
  return setUsdArray(vtarray, time);
}

//---------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------

inline bool Attribute::Impl::isAnimating() const
{
  if (valuetype == Value::InvalidType) {
    return false;
  }
  else if (sdfattr) {
    // Access time samples through SdfLayer:
    assert(sdfattr->GetLayer());
    return (sdfattr->GetLayer()->GetNumTimeSamplesForPath(sdfattr->GetPath()) > 1);
  }
  else if (usdattr) {
    return usdattr.ValueMightBeTimeVarying();
  }
  return false;
}


inline size_t Attribute::Impl::getNumTimeSamples() const
{
  if (valuetype == Value::InvalidType) {
    return 0;
  }
  else if (sdfattr) {
    // Access time samples through SdfLayer:
    assert(sdfattr->GetLayer());
    return sdfattr->GetLayer()->GetNumTimeSamplesForPath(sdfattr->GetPath());
  }
  else if (usdattr) {
    return usdattr.GetNumTimeSamples();
  }
  return 0;
}


inline size_t Attribute::Impl::getTimeSamples(fdk::TimeValueList& times) const
{
  times.clear();
  if (valuetype == Value::InvalidType) {
    return 0;
  }
  else if (sdfattr) {
    // Access time samples through SdfLayer:
    assert(sdfattr->GetLayer());
    std::set<double> timeset = sdfattr->GetLayer()->ListTimeSamplesForPath(sdfattr->GetPath());
    times.reserve(timeset.size());
    for (double time : timeset) {
      times.push_back(time);
    }
  }
  else if (usdattr) {
    usdattr.GetTimeSamples(&times);
  }
  return times.size();
}


//! Return 1 if not array type, or the size of the array.
inline size_t Attribute::Impl::getSize(const fdk::TimeValue& time) const
{
  if (valuetype == Value::InvalidType) {
    return 0;
  }
  else if (!is_array) {
    return 1;
  }

  // Array type:
  if (sdfattr) {
    // Test for time samples first:
    if (!fdk::isDefaultTimeValue(time)) {
      assert(sdfattr->GetLayer());
      PXR_NS::VtValue vt_value;
      if (sdfattr->GetLayer()->QueryTimeSample(sdfattr->GetPath(), time, &vt_value)) {
        return vt_value.GetArraySize();
      }
    }

    // Fallback to default value if it exists:
    return sdfattr->GetField(PXR_NS::SdfFieldKeys->Default).GetArraySize();
  }
  else if (usdattr && usdattr.IsValid()) {
    PXR_NS::VtValue vt_value;
    if (usdattr.Get(&vt_value, PXR_NS::UsdTimeCode(time))) {
      return vt_value.GetArraySize();
    }
  }

  return 0;
}


//!
template <typename usgT, typename usdT>
inline const usgT* Attribute::Impl::getDefaultData() const
{
  if (sdfattr) {
    const PXR_NS::VtValue v = sdfattr->GetField(PXR_NS::SdfFieldKeys->Default);
    if (v.IsEmpty()) {
      return nullptr;
    }

    if (is_array) {
      if (v.IsHolding<PXR_NS::VtArray<usdT> >()) {
        return reinterpret_cast<const usgT*>(v.UncheckedGet<PXR_NS::VtArray<usdT> >().data());
      }
    }
    else {
      if (v.IsHolding<usdT>()) {
        return reinterpret_cast<const usgT*>(&v.UncheckedGet<usdT>());
      }
    }
  }
  else if (usdattr && usdattr.IsValid()) {
  }

  return nullptr;
}

//!
template <typename usgT, typename usdT>
inline bool Attribute::Impl::getDefaultData(const usgT*& data,
                                            size_t&      size) const
{
  if (sdfattr) {
    const PXR_NS::VtValue v = sdfattr->GetField(PXR_NS::SdfFieldKeys->Default);
    if (!v.IsEmpty()) {
      if (is_array) {
        if (v.IsHolding<PXR_NS::VtArray<usdT> >()) {
          data = reinterpret_cast<const usgT*>(v.UncheckedGet<PXR_NS::VtArray<usdT> >().data());
          size = v.GetArraySize();
          return true;
        }
      }
      else {
        if (v.IsHolding<usdT>()) {
          data = reinterpret_cast<const usgT*>(&v.UncheckedGet<usdT>());
          size = 1;
          return true;
        }
      }
    }
  }
  else if (usdattr && usdattr.IsValid()) {
  }

  data = nullptr;
  size = 0;
  return false;
}


//---------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------

/* These are the value resolve states for UsdStage, we may not need to emulate
   them all for SdfAttributeSpec access:
      UsdResolveInfoSourceNone        = 0,    ///< No value
      UsdResolveInfoSourceFallback    = 1,    ///< Built-in fallback value
      UsdResolveInfoSourceDefault     = 2,    ///< Attribute default value
      UsdResolveInfoSourceTimeSamples = 3,    ///< Attribute time samples
      UsdResolveInfoSourceValueClips  = 4,    ///< Value clips
*/

inline bool Attribute::Impl::getVtValue(PXR_NS::VtValue&      vt_value,
                                        const fdk::TimeValue& time) const
{
  if (sdfattr) {
    // Test for time samples first:
    if (!fdk::isDefaultTimeValue(time)) {
      assert(sdfattr->GetLayer());
      if (sdfattr->GetLayer()->QueryTimeSample(sdfattr->GetPath(), time, &vt_value)) {
        return (!vt_value.IsEmpty());
      }
    }

    // Fallback to default value if it exists:
    vt_value = sdfattr->GetField(PXR_NS::SdfFieldKeys->Default);
    return (!vt_value.IsEmpty());
  }
  else if (usdattr && usdattr.IsValid()) {
    return usdattr.Get(&vt_value, PXR_NS::UsdTimeCode(time));
  }

  return false;
}


//!
template <typename usdT>
inline bool Attribute::Impl::getUsd(usdT&                 value,
                                    const fdk::TimeValue& time) const
{
  PXR_NS::VtValue vt_value;
  getVtValue(vt_value, time);

  if (vt_value.IsHolding<usdT>()) {
    assert(!vt_value.IsArrayValued());
    if (!vt_value.IsEmpty()) {
      value = vt_value.UncheckedGet<usdT>();
    }
    return true;
  }

  return false;
}

//!
template <typename usdT>
inline bool Attribute::Impl::getUsdArray(PXR_NS::VtArray<usdT>& value_array,
                                         const fdk::TimeValue&  time) const
{
    //std::cout << "    Attribute::Impl::getUsdArray('" << usdattr.GetName() << "')";
  PXR_NS::VtValue vt_value;
  getVtValue(vt_value, time);

  if (vt_value.IsHolding<PXR_NS::VtArray<usdT> >()) {
    assert(vt_value.IsArrayValued());
    if (!vt_value.IsEmpty()) {
      value_array = vt_value.UncheckedGet<PXR_NS::VtArray<usdT> >();
      return true;
    }
  }

  value_array.clear();

  return false;
}

//---------------------------------------------------------------------------

//!
template <typename usgT, typename usdT>
inline size_t Attribute::Impl::get(usgT&                 value,
                                   const fdk::TimeValue& time) const
{
  //if (sizeof(usgT) != sizeof(usdT)) {
  //  return false; // don't crash...
  //}
  return (getUsd(reinterpret_cast<usdT&>(value), time)) ? 1 : 0;
}

// Specialization for string handling
template <>
inline size_t Attribute::Impl::get<std::string, PXR_NS::TfToken>(std::string&          value,
                                                                const fdk::TimeValue& time) const
{
  PXR_NS::TfToken token;
  if (getUsd(token, time)) {
    value = token.GetString();
    return 1;
  }
  return 0;
}

// Specialization for Token handling
template <>
inline size_t Attribute::Impl::get<Token, PXR_NS::TfToken>(Token&           value,
                                                           const fdk::TimeValue& time) const
{
  PXR_NS::TfToken token;
  if (getUsd(token, time)) {
    value = fromTfToken(token);
    return 1;
  }
  return 0;
}

//-------------------------------------------------

//!
template <typename usgT, typename usdT>
inline size_t Attribute::Impl::getArray(std::vector<usgT>&    value_list,
                                        const fdk::TimeValue& time) const
{
  //if (sizeof(usgT) != sizeof(usdT))
  //  return false; // don't crash...

  PXR_NS::VtArray<usdT> vtarray;
  if (getUsdArray(vtarray, time)) {
    value_list.resize(vtarray.size());
    memcpy(value_list.data(), vtarray.data(), sizeof(usgT)*value_list.size());
    return value_list.size();
  }
  value_list.clear();
  return 0;
}

//!
template <>
inline size_t Attribute::Impl::getArray<std::string, PXR_NS::TfToken>(std::vector<std::string>& value_list,
                                                                     const fdk::TimeValue&     time) const
{
  PXR_NS::VtArray<PXR_NS::TfToken> vtarray;
  if (getUsdArray(vtarray, time)) {
    value_list.resize(vtarray.size());
    for (int i=0; i < value_list.size(); ++i) {
      value_list[i] = vtarray[i].GetString();
    }
    return value_list.size();
  }
  value_list.clear();
  return 0;
}

template <>
inline size_t Attribute::Impl::getArray<Token, PXR_NS::TfToken>(std::vector<Token>& value_list,
                                                                const fdk::TimeValue&    time) const
{
  PXR_NS::VtArray<PXR_NS::TfToken> vtarray;
  if (getUsdArray(vtarray, time)) {
    value_list.resize(vtarray.size());
    for (int i=0; i < value_list.size(); ++i) {
      value_list[i] = fromTfToken(vtarray[i]);
    }
    return value_list.size();
  }
  value_list.clear();
  return 0;
}


//---------------------------------------------------------------------------


//! Remove all authored values, including the default one.
inline void Attribute::Impl::clear()
{
  if (sdfattr) {
    sdfattr->ClearDefaultValue();
  }
  else if (usdattr && usdattr.IsValid()) {
    usdattr.Clear();
  }
}

//! Removes the default value, leaving time samples unaffected.
inline void Attribute::Impl::clearDefault()
{
  if (sdfattr) {
    sdfattr->ClearDefaultValue();
  }
  else if (usdattr && usdattr.IsValid()) {
    usdattr.ClearDefault();
  }
}

//! Remove all time samples from the attribute, leaving the default value unaffected.
inline void Attribute::Impl::clearAllTimeSamples()
{
  if (sdfattr) {
    return; // ignore sdfattr since it only holds default
  }
  else if (usdattr && usdattr.IsValid()) {
    std::vector<double> times;
    usdattr.GetTimeSamples(&times);
    for (size_t i=0; i < times.size(); ++i) {
      usdattr.ClearAtTime(PXR_NS::UsdTimeCode(times[i]));
    }
  }
}

//! Remove a specific time sample, ignoring default value.
inline void Attribute::Impl::clearAtTime(const fdk::TimeValue& time)
{
  if (sdfattr) {
    return; // ignore sdfattr since it only holds default
  }
  else if (usdattr && usdattr.IsValid() && !fdk::isDefaultTimeValue(time)) {
    usdattr.ClearAtTime(PXR_NS::UsdTimeCode(time));
  }
}

//! Remove all authored values, leaving the attribute as if unauthored.
inline void Attribute::Impl::block()
{
  if (sdfattr) {
    clearAllTimeSamples();
    sdfattr->SetDefaultValue(PXR_NS::VtValue(PXR_NS::SdfValueBlock()));
  }
  else if (usdattr && usdattr.IsValid()) {
    usdattr.Block();
  }
}


//---------------------------------------------------------------------------

//attrspec->GetConnectionPathList().Add(pathToAuthor);  UsdListPositionBackOfPrependList
//attrspec->GetConnectionPathList().Remove(pathToAuthor);
//attrspec->GetConnectionPathList().ClearEdits();
//attrspec->GetConnectionPathList().ClearEditsAndMakeExplicit();
//attrspec->GetConnectionPathList().GetExplicitItems() = mappedPaths;

inline bool Attribute::Impl::addConnectionTo(const Path& source)
{
  if (sdfattr) {
    // Assign connection which is a path added to a 'SdfListEditorProxy' obfuscated
    // as a 'SdfConnectionsProxy' for no bloody good reason...:
    sdfattr->GetConnectionPathList().Add(toSdfPath(source));
    return true;
  }
  else if (usdattr && usdattr.IsValid()) {
    // Underneath, UsdAttribute accesses an AttributeSpec:
    return usdattr.AddConnection(toSdfPath(source));
  }

  return true;
}

inline bool Attribute::Impl::removeConnectionTo(const Path& source)
{
  if (sdfattr) {
    // Assign connection which is a path added to a 'SdfListEditorProxy' obfuscated
    // as a 'SdfConnectionsProxy' for no bloody good reason...:
    sdfattr->GetConnectionPathList().Remove(toSdfPath(source));
    return true;
  }
  else if (usdattr && usdattr.IsValid()) {
    // Underneath, UsdAttribute accesses an AttributeSpec:
    return usdattr.RemoveConnection(toSdfPath(source));
  }

  return true;
}

inline void Attribute::Impl::clearConnections()
{
  if (sdfattr) {
    // Assign connection which is a path added to a 'SdfListEditorProxy' obfuscated
    // as a 'SdfConnectionsProxy' for no bloody good reason...:
    sdfattr->GetConnectionPathList().ClearEdits();
  }
  else if (usdattr && usdattr.IsValid()) {
    // Underneath, UsdAttribute accesses an AttributeSpec:
    usdattr.ClearConnections();
  }
}

inline Path Attribute::Impl::getConnection() const
{
  return empty_geompath;
}


//---------------------------------------------------------------------------


inline void Attribute::Impl::getFieldNames(fdk::StringList& field_names) const
{
  field_names.clear();
  if (sdfattr) {
    std::vector<PXR_NS::TfToken> field_tokens = sdfattr->ListFields();
    field_names.resize(field_tokens.size());
    for (size_t i=0; i < field_tokens.size(); ++i) {
      field_names[i] = field_tokens[i].GetString();
    }
  }
  else if (usdattr) {
    //std::map<class TfToken, VtValue>
    const PXR_NS::UsdMetadataValueMap field_map = usdattr.GetAllMetadata();
    field_names.resize(field_map.size());
    int i = 0;
    for (PXR_NS::UsdMetadataValueMap::const_iterator it=field_map.begin(); it != field_map.end(); ++it) {
      field_names[i++] = it->first.GetString();
    }
  }
}

inline bool Attribute::Impl::hasField(const PXR_NS::TfToken& field_name) const
{
  if (sdfattr) {
    return sdfattr->HasField(field_name);
  }
  else if (usdattr) {
    return usdattr.HasMetadata(field_name);
  }
  return false;
}

inline PXR_NS::VtValue Attribute::Impl::getField(const PXR_NS::TfToken& field_name) const
{
  if (sdfattr) {
    return sdfattr->GetField(field_name);
  }
  else if (usdattr) {
    PXR_NS::VtValue vtvalue;
    usdattr.GetMetadata(field_name, &vtvalue);
    return vtvalue;
  }
  return PXR_NS::VtValue();
}

inline bool Attribute::Impl::setField(const PXR_NS::TfToken& field_name,
                                      const PXR_NS::VtValue& vtvalue)
{
  if (sdfattr) {
    return sdfattr->SetField(field_name, vtvalue);
  }
  else if (usdattr) {
    return usdattr.SetMetadata(field_name, vtvalue);
  }
  return false;
}

inline bool Attribute::Impl::deleteField(const PXR_NS::TfToken& field_name)
{
  if (sdfattr) {
    return sdfattr->ClearField(field_name);
  }
  else if (usdattr) {
    // TODO: this does not appear to work...not sure why
    return usdattr.ClearMetadata(field_name);
  }
  return false;
}


//---------------------------------------------------------------------------
//---------------------------------------------------------------------------


//!
inline const std::string& Relationship::Impl::getName() const
{
  if (sdfrel) {
    return sdfrel->GetPath().GetNameToken().GetString();
  }
  else if (usdrel) {
    return usdrel.GetName().GetString();
  }
  return empty_string;
}

//!
inline const std::string& Relationship::Impl::getPath() const
{
  if (sdfrel) {
    return sdfrel->GetPath().GetString();
  }
  else if (usdrel) {
    return usdrel.GetPath().GetString();
  }
  return empty_string;
}


//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

inline const std::string& Prim::Impl::getName() const
{
  if (sdfprim)
    return sdfprim->GetPath().GetName();
  else if (usdprim)
    return usdprim.GetName().GetString();
  return empty_string;
}

inline Path Prim::Impl::getPath() const
{
  if (sdfprim)
    return fromSdfPath(sdfprim->GetPath());
  else if (usdprim)
    return fromSdfPath(usdprim.GetPath());
  return empty_geompath;
}

inline const Token
Prim::Impl::getTypeName() const
{
  if (sdfprim) {
    return fromTfToken(sdfprim->GetTypeName());
  }
  else if (usdprim) {
    return fromTfToken(usdprim.GetTypeName());
  }
  return Token();
}

inline void Prim::Impl::setActive(bool active)
{
  if (sdfprim)
    sdfprim->SetActive(active);
  else if (usdprim)
    usdprim.SetActive(active);
}

inline bool Prim::Impl::isActive() const
{
  if (sdfprim)
    return sdfprim->GetActive();
  else if (usdprim)
    return usdprim.IsActive();
  return false;
}

inline bool Prim::Impl::isDefined() const
{
  if (usdprim)
    return usdprim.IsDefined();
  return false;
}

inline bool Prim::Impl::hasPayload() const
{
    if (sdfprim)
      return sdfprim->HasPayloads();
    else if (usdprim)
      return usdprim.HasPayload();
    return false;
}

inline bool Prim::Impl::isLoaded() const
{
    if (sdfprim)
      return false;
    else if (usdprim)
      return usdprim.IsLoaded();
    return false;
}

inline void Prim::Impl::load() const
{
  if (usdprim)
    usdprim.Load();
}

inline void Prim::Impl::unload() const
{
  if (usdprim)
    usdprim.Unload();
}

inline void Prim::Impl::setKind(const Token& kind)
{
  if (sdfprim) {
    return sdfprim->SetKind(toTfToken(kind));
  }
  else if (usdprim) {
    static PXR_NS::TfToken sKindToken("kind");
    usdprim.SetMetadata(sKindToken, toTfToken(kind));
  }
}

inline Token Prim::Impl::getKind() const
{
  if (sdfprim) {
    return fromTfToken(sdfprim->GetKind());
  }
  else if (usdprim) {
    static PXR_NS::TfToken sKindToken("kind");
    PXR_NS::TfToken kind;
    usdprim.GetMetadata(sKindToken, &kind);
    return fromTfToken(kind);
  }
  return Token();
}

inline void Prim::Impl::setInstanceable(bool instanceable)
{
  if (sdfprim) {
    return sdfprim->SetInstanceable(instanceable);
  }
  else if (usdprim) {
    usdprim.SetInstanceable(instanceable);
  }
}

inline bool Prim::Impl::getInstanceable() const
{
  if (sdfprim) {
    return sdfprim->GetInstanceable();
  }
  else if (usdprim) {
    return usdprim.IsInstanceable();
  }
  return false;
}

inline bool Prim::Impl::addAppliedSchema(const Token& schemaName)
{
  const auto pxrSchemaName = toTfToken(schemaName);
  if (sdfprim) {
    auto listOp = sdfprim->GetInfo(PXR_NS::UsdTokens->apiSchemas).Get<PXR_NS::SdfTokenListOp>();

    if (listOp.IsExplicit()) {
      const auto& items = listOp.GetExplicitItems();
      if (std::find(items.begin(), items.end(), pxrSchemaName) != items.end()) {
        return true;
      }
      if (!listOp.ReplaceOperations(PXR_NS::SdfListOpTypeExplicit, items.size(), 0, {pxrSchemaName})) {
        return false;
      }
    } else {
      const auto& preItems = listOp.GetPrependedItems();
      const auto& appItems = listOp.GetAppendedItems();
      if (std::find(preItems.begin(), preItems.end(), pxrSchemaName) != preItems.end() || std::find(appItems.begin(), appItems.end(), pxrSchemaName) != appItems.end()) {
        return true;
      }
      if (!listOp.ReplaceOperations(PXR_NS::SdfListOpTypePrepended, preItems.size(), 0, {pxrSchemaName})) {
        return false;
      }
    }

    // If we got here, we edited the list op, so author it back to the spec.
    sdfprim->SetInfo(PXR_NS::UsdTokens->apiSchemas, PXR_NS::VtValue::Take(listOp));
    return true;
  }
  else if (usdprim) {
    return usdprim.AddAppliedSchema(pxrSchemaName);
  }
  return false;
}

inline void Prim::Impl::setCustomData(const std::string& key, const Value& value)
{
  if (isUsdPrim()) {
    usdprim.SetCustomDataByKey(PXR_NS::TfToken(key), value.impl()->vtvalue);
  }
  else {
    sdfprim->SetCustomData(key, value.impl()->vtvalue);
  }
}

inline bool Prim::Impl::getCustomData(const std::string& key, Value& value) const
{
  if (isUsdPrim()) {
    auto iter = usdprim.GetCustomData().find(key);
    if (iter != usdprim.GetCustomData().end()) {
      value = fromVtValue(iter->second);
      return true;
    }
  }
  else {
    auto iter = sdfprim->GetCustomData().find(key);
    if (iter != sdfprim->GetCustomData().end()) {
      value = fromVtValue(iter->second);
      return true;
    }
  }
  return false;
}

inline Value Prim::Impl::getCustomData(const std::string& key) const
{
  if (isUsdPrim()) {
    auto iter = usdprim.GetCustomData().find(key);
    if (iter != usdprim.GetCustomData().end()) {
      return fromVtValue(iter->second);
    }
  }
  else {
    auto iter = sdfprim->GetCustomData().find(key);
    if (iter != sdfprim->GetCustomData().end()) {
      return fromVtValue(iter->second);
    }
  }
  return {};
}

inline bool Prim::Impl::hasCustomData(const std::string& key) const
{
  if (isUsdPrim()) {
    auto iter = usdprim.GetCustomData().find(key);
    return iter != usdprim.GetCustomData().end();
  }
  else {
    auto iter = sdfprim->GetCustomData().find(key);
    return iter != sdfprim->GetCustomData().end();
  }
}

inline void Prim::Impl::removeCustomData(const std::string& key)
{
  if (isUsdPrim()) {
    usdprim.GetCustomData().erase(key);
  }
  else {
    sdfprim->GetCustomData().erase(key);
  }
}


//---------------------------------------------------------------------------

inline PXR_NS::SdfVariability getSdfVariability(TimeVariability variability_hint)
{
  switch (variability_hint) {
    case TimeVariability::Varying:
      // Value of property changes between time samples - interpolatable depending on data type
      return PXR_NS::SdfVariabilityVarying;

    default:
    case TimeVariability::Uniform:
      // Value is constant over time. No value or count changes between time samples (no time samples required)
      return PXR_NS::SdfVariabilityUniform;
  }
}


inline Attribute Prim::Impl::createAttr(const PXR_NS::TfToken& name,
                                        Value::Type            type,
                                        TimeVariability        variability_hint,
                                        bool                   is_custom)
{
  if (name.IsEmpty()) {
    return Attribute();
  }

  // Defining or editing?
  if (sdfprim) {
    // Does attribute already exist? If so don't create it again, Usd doesn't like that!
    const PXR_NS::SdfPath attrpath = sdfprim->GetPath().AppendProperty(name);
    PXR_NS::SdfAttributeSpecHandle sdfattr = sdfprim->GetLayer()->GetAttributeAtPath(attrpath);
    if (!sdfattr) {
      sdfattr = PXR_NS::SdfAttributeSpec::New(sdfprim,
                                              name,
                                              getSdfValueTypeName(type),
                                              getSdfVariability(variability_hint),
                                              is_custom);
      assert(sdfattr);
    }
    return Attribute(Attribute::Impl(sdfattr));
  }
  else if (usdprim) {
    PXR_NS::UsdAttribute usdattr = usdprim.GetAttribute(name);
    if (!usdattr.IsValid()) {
      usdattr = usdprim.CreateAttribute(name,
                                        getSdfValueTypeName(type),
                                        is_custom,
                                        getSdfVariability(variability_hint));
      assert(usdattr);
    }
    return Attribute(Attribute::Impl(usdattr));
  }

  return Attribute(); // failure!
}

inline Attribute Prim::Impl::createAttr(const Token&    name,
                                        Value::Type     type,
                                        TimeVariability variability_hint,
                                        bool            is_custom)
{
  return createAttr(toTfToken(name), type, variability_hint, is_custom);
}

//-------------------------------------------------

inline bool Prim::Impl::getAttr(const PXR_NS::TfToken& name,
                                Attribute&              attr)
{
  if (name.IsEmpty()) {
    return false;
  }

  // Defining or editing?
  if (sdfprim) {
    const PXR_NS::SdfPath attrpath = sdfprim->GetPath().AppendProperty(name);
    //std::cout << "path='" << attrpath.GetString() << "'" << std::endl;
    return attr.setFromImpl(Attribute::Impl(sdfprim->GetLayer()->GetAttributeAtPath(attrpath)));
  }
  else if (usdprim) {
    return attr.setFromImpl(Attribute::Impl(usdprim.GetAttribute(name)));
  }

  return false;
}

inline bool Prim::Impl::getAttr(const Token& name,
                                Attribute&   attr)
{
  return getAttr(toTfToken(name), attr);
}

//-------------------------------------------------

/*! Set a non-array attribute. Returns true on success.

    Note - if the attribute needs creating 'custom' is defaulted to false.
    If this is not desired then make sure the attribute is created
    properly first with the desired parameters.
*/
template <typename usgT, typename usdT>
inline Attribute::Impl Prim::Impl::setAttr(const PXR_NS::TfToken& name,
                                           Value::Type            type,
                                           TimeVariability        variability_hint,
                                           const usgT&            value,
                                           const fdk::TimeValue&  time)
{
  const usg::ValueTypeInfo& info = getValueTypeInfo(type);
  if (!info.isValid()) {
    return Attribute::Impl();
  }

  PXR_NS::SdfVariability variability = static_cast<PXR_NS::SdfVariability>(variability_hint);

  // Defining or editing?
  if (sdfprim) {
    // Does attribute already exist? If so don't create it again, Usd doesn't like that!
    const PXR_NS::SdfPath attrpath = sdfprim->GetPath().AppendProperty(name);
    PXR_NS::SdfAttributeSpecHandle sdfattr = sdfprim->GetLayer()->GetAttributeAtPath(attrpath);
    if (!sdfattr) {
      sdfattr = PXR_NS::SdfAttributeSpec::New(sdfprim,
                                              name,
                                              info.sdf_value_typename,
                                              variability,
                                              false/*custom*/);
    }
    assert(sdfattr);
    Attribute::Impl handle(sdfattr);
    if (handle.isValid() && handle.set<usgT, usdT>(value, time)) {
      return handle;
    }
  }
  else if (usdprim) {
    PXR_NS::UsdAttribute usdattr = usdprim.GetAttribute(name);
    if (!usdattr.IsValid()) {
      usdattr = usdprim.CreateAttribute(name,
                                        getSdfValueTypeName(type),
                                        false/*custom*/,
                                        variability);
      assert(usdattr.IsValid());
    }
    Attribute::Impl handle(usdattr);
    if (handle.isValid() && handle.set<usgT, usdT>(value, time)) {
      return handle;
    }
  }

  return Attribute::Impl();
}

template <typename usgT, typename usdT>
inline Attribute::Impl Prim::Impl::setAttr(const Token&          name,
                                           Value::Type           type,
                                           TimeVariability       variability_hint,
                                           const usgT&           value,
                                           const fdk::TimeValue& time)
{
  return setAttr(toTfToken(name), type, variability_hint, value, time);
}

//--------------------------------------------------------------------

/*! Set an array attribute. Returns true on success.

    Note - if the attribute needs creating 'custom' is defaulted to false.
    If this is not desired then make sure the attribute is created
    properly first with the desired parameters.
*/
template <typename usgT, typename usdT>
inline Attribute::Impl Prim::Impl::setArrayAttr(const PXR_NS::TfToken&   name,
                                                Value::Type              type,
                                                TimeVariability          variability_hint,
                                                const std::vector<usgT>& value_list,
                                                const fdk::TimeValue&    time)
{
  const usg::ValueTypeInfo& info = getValueTypeInfo(type);
  if (!info.isValid()) {
    return Attribute::Impl();
  }

  PXR_NS::SdfVariability variability = static_cast<PXR_NS::SdfVariability>(variability_hint);

  // Defining or editing?
  if (sdfprim) {
    // Does attribute already exist? If so don't create it again, Usd doesn't like that!
    const PXR_NS::SdfPath attrpath = sdfprim->GetPath().AppendProperty(name);
    PXR_NS::SdfAttributeSpecHandle sdfattr = sdfprim->GetLayer()->GetAttributeAtPath(attrpath);
    if (!sdfattr) {
      sdfattr = PXR_NS::SdfAttributeSpec::New(sdfprim,
                                              name,
                                              info.sdf_value_typename,
                                              variability,
                                              false/*custom*/);
    }
    assert(sdfattr);
    Attribute::Impl handle(sdfattr);
    if (handle.isValid() && handle.setArray<usgT, usdT>(value_list, time)) {
      return handle;
    }
  }
  else if (usdprim) {
    PXR_NS::UsdAttribute usdattr = usdprim.GetAttribute(name);
    if (!usdattr.IsValid()) {
      usdattr = usdprim.CreateAttribute(name,
                                        getSdfValueTypeName(type),
                                        false/*custom*/,
                                        variability);
      assert(usdattr.IsValid());
    }
    Attribute::Impl handle(usdattr);
    if (handle.isValid() && handle.setArray<usgT, usdT>(value_list, time)) {
      return handle;
    }
  }

  return Attribute::Impl();
}


//---------------------------------------------------------------------------
//---------------------------------------------------------------------------


inline Relationship Prim::Impl::createRelationship(const PXR_NS::TfToken& name,
                                                   const Path&            targetPath,
                                                   bool                   is_custom)
{
  if (name.IsEmpty()) {
    return Relationship();
  }

  // Defining or editing?
  if (sdfprim) {
    PXR_NS::SdfRelationshipSpecHandle sdfrel = PXR_NS::SdfRelationshipSpec::New(sdfprim, name, is_custom);
    // Insert the target path:
    if (sdfrel && !targetPath.empty()) {
      sdfrel->GetTargetPathList().GetExplicitItems().Insert(0, toSdfPath(targetPath));
    }
    return Relationship(Relationship::Impl(sdfrel));
  }
  else if (usdprim) {
    PXR_NS::UsdRelationship usdrel = usdprim.CreateRelationship(name, is_custom);
    // Insert the target path:
    if (usdrel && !targetPath.empty()) {
      usdrel.AddTarget(toSdfPath(targetPath));
    }
    return Relationship(Relationship::Impl(usdrel));
  }

  return Relationship(); // failure!
}


//! Retrieve a named Relationship.
inline bool Prim::Impl::getRelationship(const PXR_NS::TfToken& name,
                                        Relationship&          rel)
{
  if (name.IsEmpty()) {
    return false;
  }

  // Defining or editing?
  if (sdfprim) {
    const PXR_NS::SdfPath relpath = sdfprim->GetPath().AppendProperty(PXR_NS::TfToken(name));
    //std::cout << "path='" << relpath.GetString() << "'" << std::endl;
    return rel.setFromImpl(Relationship::Impl(sdfprim->GetLayer()->GetRelationshipAtPath(relpath)));
  }
  else if (usdprim) {
    return rel.setFromImpl(Relationship::Impl(usdprim.GetRelationship(name)));
  }

  return false;
}


inline void Prim::Impl::addReference(const std::string& identifier, const Path& path, ListPosition position)
{
  if (sdfprim) {
    auto references = sdfprim->GetReferenceList();
    const auto reference = PXR_NS::SdfReference(identifier, toSdfPath(path));
    switch (position) {
    case ListPosition::PrependFront:
      references.GetPrependedItems().Insert(0, reference);
      break;
    case ListPosition::PrependBack:
      references.GetPrependedItems().push_back(reference);
      break;
    case ListPosition::AppendFront:
      references.GetAppendedItems().Insert(0, reference);
      break;
    case ListPosition::AppendBack:
      references.GetAppendedItems().push_back(reference);
      break;
    }
  }
  else if (usdprim) {
    usdprim.GetReferences().AddReference(identifier, toSdfPath(path), PXR_NS::SdfLayerOffset(), static_cast<PXR_NS::UsdListPosition>(static_cast<int>(position)));
  }
}

inline void Prim::Impl::addPayload(const std::string& identifier, const Path& path, ListPosition position)
{
  if (sdfprim) {
    auto references = sdfprim->GetPayloadList();
    const auto payload = PXR_NS::SdfPayload(identifier, toSdfPath(path));
    switch (position) {
    case ListPosition::PrependFront:
      references.GetPrependedItems().Insert(0, payload);
      break;
    case ListPosition::PrependBack:
      references.GetPrependedItems().push_back(payload);
      break;
    case ListPosition::AppendFront:
      references.GetAppendedItems().Insert(0, payload);
      break;
    case ListPosition::AppendBack:
      references.GetAppendedItems().push_back(payload);
      break;
    }
  }
  else if (usdprim) {
    usdprim.GetPayloads().AddPayload(identifier, toSdfPath(path), PXR_NS::SdfLayerOffset(), static_cast<PXR_NS::UsdListPosition>(static_cast<int>(position)));
  }
}

inline void Prim::Impl::addVariantSet(const std::string& name, ListPosition position)
{
  if (sdfprim) {
    auto variantSets = sdfprim->GetVariantSetNameList();
    switch (position) {
    case ListPosition::PrependFront:
      variantSets.GetPrependedItems().Insert(0, name);
      break;
    case ListPosition::PrependBack:
      variantSets.GetPrependedItems().push_back(name);
      break;
    case ListPosition::AppendFront:
      variantSets.GetAppendedItems().Insert(0, name);
      break;
    case ListPosition::AppendBack:
      variantSets.GetAppendedItems().push_back(name);
      break;
    }
  }
  else if (usdprim) {
    usdprim.GetVariantSets().AddVariantSet(name, static_cast<PXR_NS::UsdListPosition>(static_cast<int>(position)));
  }
}

inline void Prim::Impl::addVariant(const std::string& variantSetName, const std::string& variantName)
{
  if (sdfprim) {
    auto variantSets = sdfprim->GetVariantSets();
    auto it = variantSets.find(variantSetName);
    if (it != variantSets.end()) {
      const PXR_NS::SdfVariantSetSpecHandle& variantSet = it->second;
      PXR_NS::SdfVariantSpec::New(variantSet, variantName);
    }
  }
  else if (usdprim) {
    usdprim.GetVariantSet(variantSetName).AddVariant(variantName);
  }
}

inline void Prim::Impl::setVariantSelection(const std::string& variantSetName, const std::string& selection)
{
  if (sdfprim) {
    sdfprim->SetVariantSelection(variantSetName, selection);
  }
  else if (usdprim) {
    usdprim.GetVariantSet(variantSetName).SetVariantSelection(selection);
  }
}

inline std::vector<std::string> Prim::Impl::getVariantSets() const
{
  std::vector<std::string> result;
  if (sdfprim) {
    for (const auto it : sdfprim->GetVariantSets()) {
      const PXR_NS::SdfVariantSetSpecHandle& variantSet = it.second;
      result.push_back(variantSet.GetSpec().GetName());
    }
  }
  else if (usdprim) {
    result = usdprim.GetVariantSets().GetNames();
  }
  return result;
}

inline std::vector<std::string> Prim::Impl::getVariants(const std::string& variantSetName) const
{
  if (sdfprim) {
    return sdfprim->GetVariantNames(variantSetName);
  }
  else if (usdprim) {
    return usdprim.GetVariantSet(variantSetName).GetVariantNames();
  }
  return std::vector<std::string>();
}


} // namespace usg

#endif

