// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/base/Token.h"

#include "usg/private/MemoryPoolWrapperImpl.h"

#include "fdk/base/Hash.h"
#include "fdk/base/MemoryPool.h"

#include "pxr/base/tf/token.h"

namespace usg {
  namespace {
    MemoryPoolWrapperImpl<PXR_NS::TfToken, Token, Token::Handle>& wrapper()
    {
      static MemoryPoolWrapperImpl<PXR_NS::TfToken, Token, Token::Handle> sWrapper;
      return sWrapper;
    }

    const PXR_NS::TfToken& unwrap(const MemoryPoolWrapperImpl<PXR_NS::TfToken, Token, Token::Handle>::MemoryPoolHandle& h)
    {
      return wrapper().unwrap(h);
    }

    Token wrap(const PXR_NS::TfToken& token)
    {
      return wrapper().wrap(token);
    }
  }

  Token::Token()
    : _handle(wrapper().allocate())
  {
  }

  Token::Token(const char* text)
    : _handle(wrapper().allocate(text))
  {
  }

  Token::Token(const std::string& text)
    : _handle(wrapper().allocate(text.c_str()))
  {
  }

  Token::Token(const Token::Handle& handle)
    : _handle(wrapper().copyAllocate(handle))
  {
  }

  Token::~Token()
  {
    wrapper().free(_handle);
  }

  Token::Token(Token&& token) noexcept
  {
    _handle = token._handle;
    token._handle = fdk::MemoryPool<PXR_NS::TfToken>::nullhandle;
  }

  Token& Token::operator=(Token&& rhs) noexcept
  {
    if (this != &rhs) {
      _handle = rhs._handle;
      rhs._handle = fdk::MemoryPool<PXR_NS::TfToken>::nullhandle;
    }
    return *this;
  }

  Token::Token(const Token& rhs)
    : _handle(wrapper().copyAllocate(rhs._handle))
  {
  }

  Token& Token::operator=(const Token& rhs)
  {
    if (this != &rhs) {
      _handle = wrapper().copyAllocate(rhs._handle);
    }
    return *this;
  }

  const char* Token::text() const
  {
    return unwrap(_handle).GetText();
  }

  const std::string& Token::asString() const
  {
    return unwrap(_handle).GetString();
  }

  bool Token::empty() const
  {
    return unwrap(_handle).IsEmpty();
  }

  bool Token::operator==(const Token& rhs) const
  {
    return unwrap(_handle) == unwrap(rhs._handle);
  }

  bool Token::operator<(const Token& rhs) const
  {
    return unwrap(_handle) < unwrap(rhs._handle);
  }

  Token::Handle& Token::getToken(int usd_version)
  {
    return wrapper().getWrappedObject(_handle);
  }

  const Token::Handle& Token::getToken(int usd_version) const
  {
    return wrapper().getWrappedObject(_handle);
  }

  uint64_t Token::hash() const
  {
    return unwrap(_handle).Hash();
  }

}
