// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/geom/NotificationDelegate.h"

#include <pxr/base/tf/diagnosticMgr.h>

#include <iostream>

namespace {
  usg::NotificationDelegate* gNotificationDelegate;

  bool sPrintErrors = true; // 

  class ErrorDelegate : public PXR_NS::TfDiagnosticMgr::Delegate {
  public:
    ErrorDelegate();
    ~ErrorDelegate() override = default;

    void IssueError(const PXR_NS::TfError& err) override;
    void IssueFatalError(const PXR_NS::TfCallContext& context, const std::string& msg) override;
    void IssueStatus(const PXR_NS::TfStatus& status) override;
    void IssueWarning(const PXR_NS::TfWarning& warning) override;
  };

  ErrorDelegate::ErrorDelegate()
  {
    PXR_NS::TfDiagnosticMgr::GetInstance().AddDelegate(this);
  }

  void ErrorDelegate::IssueError(const PXR_NS::TfError& err)
  {
    if (gNotificationDelegate) {
      gNotificationDelegate->error(err.GetCommentary());
    }
    if (sPrintErrors) {
      std::cerr << "Error: " << err.GetCommentary() << std::endl;
    }
  }

  void ErrorDelegate::IssueFatalError(const PXR_NS::TfCallContext& context, const std::string& msg)
  {
    if (gNotificationDelegate) {
      gNotificationDelegate->fatalError(msg);
    }
    if (sPrintErrors) {
      std::cerr << "Fatal Error: " << msg << std::endl;
    }
  }

  void ErrorDelegate::IssueStatus(const PXR_NS::TfStatus& status)
  {
    if (gNotificationDelegate) {
      gNotificationDelegate->status(status.GetDiagnosticCodeAsString());
    }
    if (sPrintErrors) {
      std::cout << "Status: " << status.GetDiagnosticCodeAsString() << std::endl;
    }
  }

  void ErrorDelegate::IssueWarning(const PXR_NS::TfWarning& warning)
  {
    if (gNotificationDelegate) {
      gNotificationDelegate->warning(warning.GetCommentary());
    }
    if (sPrintErrors) {
      std::cerr << "Warning: " << warning.GetCommentary() << std::endl;
    }
  }

  ErrorDelegate sErrorDelegate;
}

namespace usg
{
  USG_API void setNotificationDelegate(NotificationDelegate* notificationDelegate)
  {
    gNotificationDelegate = notificationDelegate;
  }
}

