// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/geom/Python.h"

#include "usg/geom/Stage.h"
#include "usg/geom/Prim.h"

#include "pxr/pxr.h"
#ifdef PXR_PYTHON_SUPPORT_ENABLED
#define Boost_NAMESPACE foundryboost
#include "pxr/base/tf/pyPtrHelpers.h"
#include "pxr/usd/sdf/layer.h"
#include "pxr/usd/usd/stage.h"
#endif

namespace usg
{
  USG_API PyObject* getPythonObject(const StageRef& stage)
  {
#ifdef PXR_PYTHON_SUPPORT_ENABLED
    if (stage) {
      auto handle = stage->getUsdStageRefPtr(usg::usdAPIVersion());
      if (handle) {
        PXR_NS::UsdStageWeakPtr weakPtr = *reinterpret_cast<PXR_NS::UsdStageRefPtr*>(handle);
        const auto result = PXR_NS::TfMakePyPtr<PXR_NS::UsdStageWeakPtr>::Execute(weakPtr);
        if (result.second) {
          return result.first;
        }
      }
    }
#endif
    return nullptr;
  }

  USG_API PyObject* getPythonObject(const LayerRef& layer)
  {
#ifdef PXR_PYTHON_SUPPORT_ENABLED
    if (layer) {
      auto handle = layer->getSdfLayerRefPtr(usg::usdAPIVersion());
      if (handle) {
        PXR_NS::SdfLayerHandle weakPtr = *reinterpret_cast<PXR_NS::SdfLayerRefPtr*>(handle);
        const auto result = PXR_NS::TfMakePyPtr<PXR_NS::SdfLayerHandle>::Execute(weakPtr);
        if (result.second) {
          return result.first;
        }
      }
    }
#endif
    return nullptr;
  }

}
