// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/geom/Relationship.h"

#include "usg/private/UsdImpl.h"

namespace usg
{

extern std::string empty_string;
extern Path    empty_geompath;


/*! Default ctor is invalid.
*/
Relationship::Relationship() :
  _pImpl(nullptr)
{
  //
}


Relationship::Relationship(const Relationship& b) :
  _pImpl(new Relationship::Impl(*b._pImpl.get()))
{
  //
}
Relationship& Relationship::operator = (const Relationship& b)
{
  if (_pImpl.get() != b._pImpl.get())
    _pImpl.reset(new Relationship::Impl(*b._pImpl.get()));
  return *this;
}


Relationship::Relationship(const Relationship::Impl& b) :
  _pImpl(new Relationship::Impl(b))
{
  //
}
Relationship& Relationship::operator = (const Relationship::Impl& b)
{
  if (_pImpl.get() != &b)
    _pImpl.reset(new Relationship::Impl(b));
  return *this;
}


Relationship::~Relationship()
{
  //
}

bool
Relationship::setFromImpl(const Relationship::Impl& b)
{
  *this = b;
  return _pImpl->isValid();
}

bool Relationship::isValid() const
{
  return (_pImpl && _pImpl->isValid());
}


Relationship::RelationshipSpecHandle* Relationship::getSdfRelationshipSpec(int usd_version)
{
  return (_pImpl && usdVersionMatchesUsg(usd_version)) ?
            reinterpret_cast<Relationship::RelationshipSpecHandle*>(&_pImpl->sdfrel) :
            nullptr;
}
const Relationship::RelationshipSpecHandle* Relationship::getSdfRelationshipSpec(int usd_version) const
{
  return (_pImpl && usdVersionMatchesUsg(usd_version)) ?
            reinterpret_cast<const Relationship::RelationshipSpecHandle*>(&_pImpl->sdfrel) :
            nullptr;
}

Relationship::UsdRelationshipHandle* Relationship::getUsdRelationship(int usd_version)
{
  return (_pImpl && usdVersionMatchesUsg(usd_version)) ?
            reinterpret_cast<Relationship::UsdRelationshipHandle*>(&_pImpl->usdrel) :
            nullptr;
}
const Relationship::UsdRelationshipHandle* Relationship::getUsdRelationship(int usd_version) const
{
  return (_pImpl && usdVersionMatchesUsg(usd_version)) ?
            reinterpret_cast<const Relationship::UsdRelationshipHandle*>(&_pImpl->usdrel) :
            nullptr;
}


const std::string& Relationship::getName() const { return (_pImpl) ? _pImpl->getName() : empty_string; }
Path               Relationship::getPath() const { return (_pImpl) ? _pImpl->getPath() : empty_geompath; }
void               Relationship::setTarget(const Path& targetPath) { if (_pImpl) _pImpl->setTarget(targetPath); }
Path               Relationship::getTarget() const { return (_pImpl) ? _pImpl->getTarget() : empty_geompath; }
void               Relationship::setTargets(const PathArray& targetPaths) { if (_pImpl) _pImpl->setTargets(targetPaths); }
PathArray          Relationship::getTargets() const { return (_pImpl) ? _pImpl->getTargets() : PathArray(); }


std::ostream& operator << (std::ostream& o, const Relationship& rel)
{
  const Relationship::Impl* handle = rel.impl();
  if (handle)
  {
    o << handle->getTarget();
  }
  return o;
}

#if 0
// TODO: move these back to Relationship header and implement the ones we want to keep:

#if 0
    //! USD - Make the authoring layer's opinion of the targets list explicit, and set exactly to \p targets.
    bool             setTargets(const PathArray& targets) const;


    //! SDF - Returns the relationship's target path list editor.
    SdfTargetsProxy  getTargetList() const;
    //! USD - Compose this relationship's targets and fill \p targets with the result.
    bool             getTargets(PathArray* targets) const;



    //! SDF - Returns true if the relationship has any target paths.
    bool             hasTargetList() const;
    //! SDF - Clears the list of target paths on this relationship.
    void             clearTargetList() const;
    //! USD - Remove all opinions about the target list from the current edit target.
    bool             clearTargets(bool removeSpec) const;


    //! SDF - Updates the specified target path.
    void             replaceTarget(const Path& oldPath,
                                   const Path& newPath);
    //! USD - 
    bool             addTarget(const Path& target,
                               UsdListPosition position = UsdListPositionBackOfPrependList) const;


    //! SDF - Removes the specified target path.
    void             removeTarget(const Path& path,
                                  bool            preserveTargetOrder = false);
    //! USD - Removes \p target from the list of targets.
    bool             removeTarget(const Path& target) const;
#endif


  public:
    //------------------------------------------------------
    // Misc:
    //------------------------------------------------------

#if 0
    //! SDF - Get whether loading the target of this relationship is necessary to load the prim we're attached to
    bool  getNoLoadHint(void) const;
    //! SDF - Set whether loading the target of this relationship is necessary to load the prim we're attached to
    void  setNoLoadHint(bool noload);

    //! USD - Returns true if any target path opinions have been authored. 
    bool  hasAuthoredTargets() const;
    //! USD - Compose this relationship's \em ultimate targets, taking into account "relationship forwarding", and fill \p targets with the result.
    bool  getForwardedTargets(PathArray* targets) const;
#endif

#endif

} // namespace usg
