// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/geom/ShaderProperty.h"

#include "usg/private/UsdImpl.h"

namespace usg
{


/*!
*/
ShaderProperty::ShaderProperty(int          inputOrOutput,
                               Value::Type  type,
                               const char*  name,
                               const Value& value,
                               size_t       arraySize) :
  SchemaProperty(inputOrOutput, type, name, value, arraySize)
{
  //
}

/*!
*/
ShaderProperty::ShaderProperty(int                     inputOrOutput,
                               Value::Type             type,
                               const char*             name,
                               const fdk::KeyValueMap& metadata) :
  SchemaProperty(inputOrOutput, type, name, metadata)
{
  //
}

/*!
*/
ShaderProperty::ShaderProperty(int                     inputOrOutput,
                               Value::Type             type,
                               const char*             name,
                               const Value&            value,
                               size_t                  arraySize,
                               const fdk::KeyValueMap& metadata,
                               const fdk::KeyValueMap& hints,
                               const fdk::StringList&  options) :
  SchemaProperty(inputOrOutput, type, name, value, arraySize, metadata, hints, options)
{
  //
}


ShaderProperty::ShaderProperty(const ShaderProperty& b) :
  SchemaProperty(b)
{
  //
}

ShaderProperty::ShaderProperty(const ShaderPropertyHandle& handle) :
  SchemaProperty(handle)
{
  //
}


std::string         ShaderProperty::getInfoString() const { return _pImpl->getProperty()->GetInfoString(); }
const std::string&  ShaderProperty::getLabel()      const { return _pImpl->getProperty()->GetLabel().GetString(); }
std::string         ShaderProperty::getHelp()       const { return _pImpl->getProperty()->GetHelp(); }
const std::string&  ShaderProperty::getPage()       const { return _pImpl->getProperty()->GetPage().GetString(); }
const std::string&  ShaderProperty::getWidget()     const { return _pImpl->getProperty()->GetWidget().GetString(); }

fdk::KeyValueMap
ShaderProperty::getHints() const
{
  const auto& ndrHints = _pImpl->getProperty()->GetHints();
  fdk::KeyValueMap hints;
  hints.reserve(ndrHints.size());
  for (const auto& hint : ndrHints) {
    hints.emplace(hint.first.GetText(), hint.second);
  }
  return hints;
}

fdk::StringList
ShaderProperty::getOptions() const
{
  fdk::StringList options;
  const auto &ndrOptions = _pImpl->getProperty()->GetOptions();
  options.reserve(ndrOptions.size());
  for (const auto &option : ndrOptions) {
    options.push_back(option.first.GetText());
  }
  return options;
}


} // namespace usg
