// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "usg/geom/XformCache.h"
#include "usg/private/UsdImpl.h"

namespace usg {

  class XformCacheImpl : public PXR_NS::UsdGeomXformCache {
  public:
    XformCacheImpl(fdk::TimeValue time = fdk::defaultTimeValue())
      : PXR_NS::UsdGeomXformCache(static_cast<PXR_NS::UsdTimeCode>(time))
    {
    }
  };

  XformCache::XformCache(fdk::TimeValue time)
  {
    _pImpl.reset(new XformCacheImpl(time));
  }

  XformCache::~XformCache()
  {
  }

  fdk::Mat4d XformCache::getLocalToWorldTransform(const Prim &prim)
  {
    return fdk::Mat4d(_pImpl->GetLocalToWorldTransform(prim.impl()->usdprim).GetArray());
  }

  fdk::Mat4d XformCache::getParentToWorldTransform(const Prim &prim)
  {
    return fdk::Mat4d(_pImpl->GetParentToWorldTransform(prim.impl()->usdprim).GetArray());
  }

  fdk::Mat4d XformCache::getLocalTransformation (const Prim &prim, bool *resetsXformStack)
  {
    return fdk::Mat4d(_pImpl->GetLocalTransformation(prim.impl()->usdprim, resetsXformStack).GetArray());
  }

  fdk::Mat4d XformCache::computeRelativeTransform (const Prim &prim, const Prim &ancestor, bool *resetsXformStack)
  {
    return fdk::Mat4d(_pImpl->ComputeRelativeTransform(prim.impl()->usdprim, ancestor.impl()->usdprim, resetsXformStack).GetArray());
  }

}
