// Copyright (c) 2022 The Foundry Visionmongers Ltd. All Rights Reserved.

// This is automatically-generated code.
// You may add custom code between the --(BEGIN CUSTOM CODE)-- and --(END CUSTOM CODE)-- markers.
// Code added anywhere else will be destroyed when the code is generated.

#include "usg/geom/XformablePrim.h"
#include "usg/geom/Layer.h"
#include "usg/geom/Stage.h"

namespace usg {

  XformablePrim::XformablePrim() : ImageablePrim() {}
  XformablePrim::XformablePrim(const Prim& prim) : ImageablePrim(prim) {
    validate();
  }
  XformablePrim::~XformablePrim() {}

  bool XformablePrim::validate() {
    if (!isValidPrim(*this)) {
      *this = XformablePrim();
      return false;
    }
    return true;
  }

  bool XformablePrim::isValidPrim(const Prim& prim) {
    return prim.isA(GeomTokens.types.Xformable);
  }
  

  XformablePrim XformablePrim::overrideInLayer(const LayerRef& layer, const Prim& prim)
  {
    return layer ? XformablePrim(layer->overridePrim(prim)) : XformablePrim();
  }

  XformablePrim XformablePrim::getInStage(const StageRef& stage, const Path&  path)
  {
    return XformablePrim(stage->getPrimAtPath(path));
  }

  Attribute XformablePrim::createXformOpOrderAttr(const Value& defaultValue) {
    return createAttr(GeomTokens.xformOpOrder, Value::TokenArray, false, TimeVariability::Uniform, defaultValue);
  }
  Attribute XformablePrim::getXformOpOrderAttr() const {
    return getAttr(GeomTokens.xformOpOrder);
  }
  void XformablePrim::setXformOpOrder(const TokenArray& value, fdk::TimeValue time) {
    createXformOpOrderAttr().setValue(value, time);
  }
  TokenArray XformablePrim::getXformOpOrder(fdk::TimeValue time) const {
    TokenArray v; getAttr(GeomTokens.xformOpOrder).getValue<TokenArray>(v, time); return v;
  }


  // --(BEGIN CUSTOM CODE)--
}

#include "usg/geom/ScopePrim.h"
#include "usg/private/UsdImpl.h"

namespace usg {

  /*static*/ std::string
  XformablePrim::getXformOpAttrName(const std::string& idName)
  {
    std::string opName("xformOp:transform");
    if (!idName.empty()) {
      opName += ":" + idName;
    }
    return opName;
  }

  /*static*/ bool
  XformablePrim::canConcatenateTransform(const Prim& prim)
  {
    return (prim.isValid()       &&
            !prim.isPseudoRoot() &&
            (isValidPrim(prim) || ScopePrim::isValidPrim(prim)));
  }

  /*static*/ size_t
  XformablePrim::getConcatenatedXformOpTimeSamples(const Prim&        prim,
                                                   fdk::TimeValueSet& times)
  {
    if (!canConcatenateTransform(prim)) {
      times.clear();
      return 0; // at top, stop
    }
    getConcatenatedXformOpTimeSamples(prim.getParent(), times); // walk up

    XformablePrim xformable = XformablePrim(prim);
    if (xformable.validate()) {
      // Get the Xform key range:
      fdk::TimeValueList xform_times;
      if (xformable.getTimeSamples(xform_times)) {
        for (const fdk::TimeValue& time : xform_times) {
          times.insert(time);
        }
      }
    }
    return times.size();
  }

  /*static*/ fdk::Mat4d
  XformablePrim::getLocalMatrixAtPrim(const Prim&           prim,
                                      const fdk::TimeValue& time)
  {
    if (!prim.isValid() || !prim.impl()->usdprim) {
      return fdk::Mat4d(1.0);
    }

    // Use the Xform cache system rather than direct xform access on the schema:
    PXR_NS::UsdGeomXformCache xform_cache( (PXR_NS::UsdTimeCode(time)) );

    bool resets_xform_stack = false;
    return fdk::Mat4d(xform_cache.GetLocalTransformation(prim.impl()->usdprim, &resets_xform_stack).GetArray());
  }

  /*static*/ fdk::Mat4d
  XformablePrim::getConcatenatedMatrixAtPrim(const Prim&           prim,
                                             const fdk::TimeValue& time)
  {
    if (!prim.isValid() || !prim.impl()->usdprim) {
      return fdk::Mat4d(1.0);
    }

    // Use the Xform cache system rather than direct xform access on the schema:
    PXR_NS::UsdGeomXformCache xform_cache( (PXR_NS::UsdTimeCode(time)) );

    return fdk::Mat4d(xform_cache.GetLocalToWorldTransform(prim.impl()->usdprim).GetArray());
  }

  /*static*/ void
  XformablePrim::getConcatenatedMatricesAtPrim(const Prim&               prim,
                                               const fdk::TimeValueList& times,
                                               Mat4dArray&               matrices)
  {
    matrices.clear();

    if (!prim.isValid() || !prim.impl()->usdprim) {
      return;
    }

    const size_t nSamples = times.size();
    if (nSamples == 0) {
      // Uniform:
      PXR_NS::UsdGeomXformCache xform_cache(PXR_NS::UsdTimeCode::Default());
      matrices.resize(1);
      matrices[0] = fdk::Mat4d(xform_cache.GetLocalToWorldTransform(prim.impl()->usdprim).GetArray());
    }
    else {
      // Animated:
      matrices.resize(nSamples);
      for (size_t i=0; i < nSamples; ++i) {
        PXR_NS::UsdGeomXformCache xform_cache( (PXR_NS::UsdTimeCode(times[i])) );
        matrices[i] = fdk::Mat4d(xform_cache.GetLocalToWorldTransform(prim.impl()->usdprim).GetArray());
      }
    }
  }


  //---------------------------------------------------------------------------


  fdk::Mat4d
  XformablePrim::getLocalXform(const fdk::TimeValue& time) const
  {
    if (!_pImpl || !_pImpl->usdprim) {
      return fdk::Mat4d(1.0);
    }

    // Use the Xform cache system rather than direct xform access on the schema:
    PXR_NS::UsdGeomXformCache xform_cache( (PXR_NS::UsdTimeCode(time)) );

    bool resets_xform_stack = false;
    return fdk::Mat4d(xform_cache.GetLocalTransformation(_pImpl->usdprim, &resets_xform_stack).GetArray());
  }

  fdk::Mat4d
  XformablePrim::getConcatenatedXform(const fdk::TimeValue& time) const
  {
    if (!_pImpl || !_pImpl->usdprim) {
      return fdk::Mat4d(1.0);
    }

    // Use the Xform cache system rather than direct xform access on the schema:
    PXR_NS::UsdGeomXformCache xform_cache( (PXR_NS::UsdTimeCode(time)) );

    return fdk::Mat4d(xform_cache.GetLocalToWorldTransform(_pImpl->usdprim).GetArray());
  }

  bool
  XformablePrim::getTimeSamples(fdk::TimeValueList& times) const
  {
    if (!_pImpl || !_pImpl->usdprim) {
      return false;
    }
    return PXR_NS::UsdGeomXformable(_pImpl->usdprim).GetTimeSamples(&times);
  }


  //---------------------------------------------------------------------------

  /*
      //----------------------------------

      UsdGeomXformOp AddXformOp(UsdGeomXformOp::Type const opType,
                                UsdGeomXformOp::Precision const
                                precision=UsdGeomXformOp::PrecisionDouble,
                                TfToken const &opSuffix = TfToken(),
                                bool isInverseOp=false) const;


  */

  void
  XformablePrim::setXformTo(const std::string&    id_name,
                            const fdk::Mat4d&     xform,
                            const fdk::TimeValue& time)
  {
    if (!_pImpl) {
      return;
    }

    const std::string op_name = getXformOpAttrName(id_name);
    Attribute transformOpAttr = _pImpl->createAttr(op_name, Value::Matrix4d);
    if (!transformOpAttr) {
      std::cerr << getName() << "::setXformOp() error, cannot create attribute '" << op_name << "', ";
      std::cerr << "this is likely due to a coding error." << std::endl;
      return;
    }
    // The reason for using the VtValue API here is to avoid problems with trying to instantiate
    // templates before their specialization
    transformOpAttr.impl()->setVtValue(PXR_NS::VtValue(reinterpret_cast<const PXR_NS::GfMatrix4d&>(xform)), time);

    if (Attribute xformOpOrderAttr = _pImpl->createAttr(GeomTokens.xformOpOrder, Value::TokenArray)) {
      PXR_NS::VtArray<PXR_NS::TfToken> tokens {PXR_NS::TfToken(op_name)};
      // xformOoOrder is always uniform!
      xformOpOrderAttr.impl()->setVtValue(PXR_NS::VtValue(tokens), fdk::defaultTimeValue());
    }
  }

  void
  XformablePrim::addXformOp(const std::string&    id_name,
                            const fdk::Mat4d&     xform,
                            const fdk::TimeValue& time,
                            bool                  prepend)
  {
    if (!_pImpl) {
      return;
    }

    // xformOpOrder attrib should always exist and have a value if there's an xform applied.
    // If empty then assume there's no existing xform:
    Attribute xformOpOrderAttr;
    getAttr(GeomTokens.xformOpOrder, xformOpOrderAttr);
    if (!xformOpOrderAttr/* || xformOpOrderAttr.isEmpty()*/) {
      // No existing xform ops, no need to append:
      setXformTo(id_name, xform, time);
      return;
    }

    // Append/prepend to xform op list:
    const std::string new_name = getXformOpAttrName(id_name);
    const PXR_NS::TfToken op_name(new_name);

    // Add XformOp:
    Attribute transformOpAttr = _pImpl->createAttr(op_name, Value::Matrix4d);
    if (!transformOpAttr) {
      std::cerr << getName() << "::addXformOp() error, cannot create attribute '" << op_name << "', ";
      std::cerr << "this is likely due to a coding error." << std::endl;
      return;
    }
    // The reason for using the VtValue API here is to avoid problems with trying to instantiate
    // templates before their specialization
    transformOpAttr.impl()->setVtValue(PXR_NS::VtValue(reinterpret_cast<const PXR_NS::GfMatrix4d&>(xform)), time);

    // Append or prepend the new XformOp to the xformOpOrder list:
    {
      Value xformOps;
      if (!xformOpOrderAttr.getValue(xformOps, fdk::defaultTimeValue())) {
        // No existing xform ops, no need to append:
        setXformTo(id_name, xform, time);
        return;
      }
      PXR_NS::VtValue& xformOpsVtValue = xformOps.impl()->vtvalue;
      if (!xformOpsVtValue.IsHolding<PXR_NS::VtArray<PXR_NS::TfToken>>()) {
        std::cerr << getName() << "::addXformOp() error, invalid Value type for attribute 'xformOpOrder', ";
        std::cerr << "this is likely due to a coding error." << std::endl;
        return;
      }

      PXR_NS::VtArray<PXR_NS::TfToken> xformOpOrder = xformOpsVtValue.UncheckedGet<PXR_NS::VtArray<PXR_NS::TfToken>>();
      xformOpOrder.resize(xformOpOrder.size() + 1);
      if (prepend) {
        for (std::size_t i = xformOpOrder.size() - 1; i > 0; --i) {
          xformOpOrder[i] = xformOpOrder[i - 1];
        }
        xformOpOrder[0] = op_name;
      }
      else {
        xformOpOrder[xformOpOrder.size() - 1] = op_name;
      }
      xformOpsVtValue = xformOpOrder;
      xformOpOrderAttr.impl()->setVtValue(xformOpsVtValue, fdk::defaultTimeValue());
    }
  }

  void
  XformablePrim::clearAllXformOps()
  {
    if (!_pImpl) {
      return;
    }

    if (_pImpl->sdfprim) {
      for (const std::string& xformName : getOrderedXformOps()) {
        if (auto propHandle = _pImpl->sdfprim->GetProperties()[xformName]) {
          _pImpl->sdfprim->RemoveProperty(propHandle);
        }
      }
      if (auto propHandle = _pImpl->sdfprim->GetProperties()[GeomTokens.xformOpOrder.asString()]) {
        _pImpl->sdfprim->RemoveProperty(propHandle);
      }
    }
  }

  fdk::StringList
  XformablePrim::getOrderedXformOps() const
  {
    fdk::StringList names;
    if (!_pImpl) {
      return names;
    }

    Attribute xformOpOrderAttr;
    if (getAttr(GeomTokens.xformOpOrder, xformOpOrderAttr)) {
      Value xformOps;
      if (!xformOpOrderAttr.getValue(xformOps, fdk::defaultTimeValue())) {
        return names;
      }
      PXR_NS::VtValue vtValue = xformOps.impl()->vtvalue;

      if (vtValue.IsHolding<PXR_NS::VtArray<PXR_NS::TfToken>>()) {
        PXR_NS::VtArray<PXR_NS::TfToken> xformOpOrder = vtValue.UncheckedGet<PXR_NS::VtArray<PXR_NS::TfToken>>();
        for (const auto& xformOp : xformOpOrder) {
          names.emplace_back(xformOp.GetText());
        }
      }
    }

    return names;
  }

  namespace
  {
    inline bool isInverse(const std::string& xformOpName)
    {
      return xformOpName.find("!invert!") == 0;
    }
  }

  void XformablePrim::copyXformOps(const XformablePrim& src_xformable, fdk::TimeValue time)
  {
    for (const std::string& xformName : src_xformable.getOrderedXformOps()) {
      if (!isInverse(xformName)) {
        copyAttr(src_xformable, xformName, time);
      }
    }
    // xformOpOrder is always uniform, no time samples!
    copyAttr(src_xformable, GeomTokens.xformOpOrder, fdk::defaultTimeValue());
  }
  // --(END CUSTOM CODE)--
}
